#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Generate synthetic spectra that can be used to test specex.rrspecex module.

Copyright (C) 2022  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
import os
import json
import warnings
import numpy as np
from astropy.io import fits
from astropy import units
from scipy.signal import savgol_filter
from astropy.utils.exceptions import AstropyWarning


from test import TEST_DATA_PATH
from test import make_synt_cube

from specex.utils import get_sdss_spectral_templates, get_sdss_template_data


def fake_spectrum_fits(obj_id, wave_range, template, z, wave_step=1):
    """
    Generate a fits HDUList containing a synthetic spectrum.

    Parameters
    ----------
    obj_id : integer
        A unique id for this synthetic object. It's reccomended to use positive
        increasing integers when creating multiple spectra that will be used
        at the same time withrrspecex.
    wave_range : tuple or numpy.ndarray
        The range of wavelenght, in Angstrom, the sythetic spectrum should
        cover. Only maximum and minimum values of this parameter are used.
    template : dict
        A spectral template dictionary as returned by
        specex.utils.get_sdss_template_data()
    z : float
        Redshift of the spectrum.
    wave_step : float, optional
        Resolution of the spectrum in Angstrom. The default is 1.

    Returns
    -------
    hdul : astropy.io.fits.HDUList
        A fits HDU list containing three extensions:
            - EXT 0 : PrimaryHDU
                This extension contains only an header with information about
                the spectrum
            - EXT 1 : ImageHDU
                This extension contains the spectrum itself. The fulxes are
                expressed in 10**(-20)*erg/s/cm**2/Angstrom'.
                The header of this extension contains WCS data that can be
                used to compute the wavelength for each pixel of the spectrum.
            - EXT 2 : ImageHDU
                This extension contains the variance of the spectrum.
                Values are expressed in 10**(-20)*erg/s/cm**2/Angstrom'.
                The header of this extension contains WCS data that can be
                used to compute the wavelength for each pixel of the spectrum.
    """
    wave, flux = make_synt_cube.gen_fake_spectrum(
        wave_range, template, z, wave_step
    )

    noise = np.random.standard_normal(len(flux)) * 0.1 * np.std(flux)

    spec_header = fits.header.Header()
    spec_header['CRVAL1'] = wave[0]
    spec_header['CRPIX1'] = 1.0
    spec_header['CDELT1'] = wave_step
    spec_header['CUNIT1'] = 'Angstrom'
    spec_header['BUNIT'] = '10**(-20)*erg/s/cm**2/Angstrom'
    spec_header['CTYPE1'] = 'WAVE'
    spec_header['OBJECT'] = 'SYNTHETIC'
    spec_header['PCOUNT'] = 0
    spec_header['GCOUNT'] = 1

    spec_hdu = fits.ImageHDU(data=flux+noise, header=spec_header)
    spec_hdu.name = 'SPECTRUM'

    var = 2*np.ones_like(flux)
    delta_flux = (flux[:-1] - flux[1:])**2
    delta_flux -= np.nanmin(delta_flux)
    delta_flux /= np.nanmax(delta_flux)
    var[:-1] = 2 + 100 * delta_flux

    var_hdu = fits.ImageHDU(data=var, header=spec_header)
    var_hdu.name = 'VARIANCE'

    # Smoothing the spectrum to get a crude approximation of the continuum
    smoothed_spec = savgol_filter(flux, 51, 11)

    # Subtract the smoothed spectrum to the spectrum itself to get a
    # crude estimation of the noise
    noise_spec = np.nanstd(flux - smoothed_spec)

    # Get the mean value of the spectrum
    obj_mean_spec = np.nanmean(flux)

    # Get the mean Signal to Noise ratio
    sn_spec = obj_mean_spec / np.nanmean(noise_spec)
    if np.isnan(sn_spec):
        sn_spec = 99

    apertures = [
        10 * np.random.rand() * units.arcsec,
        10 * np.random.rand() * units.arcsec,
        360 * np.random.rand() * units.deg
    ]

    apertures_info = [
        x.to_string() for x in apertures
    ]

    prim_hdu = fits.PrimaryHDU()
    prim_hdu.header['OBJ_Z'] = z
    prim_hdu.header['NPIX'] = 10
    prim_hdu.header['SN'] = sn_spec
    prim_hdu.header['ID'] = obj_id
    prim_hdu.header['RA'] = '50.0'
    prim_hdu.header['DEC'] = '50.0'
    prim_hdu.header['Z'] = z
    prim_hdu.header['EXT_MODE'] = 'kron-ellipse'
    prim_hdu.header['EXT_APER'] = json.dumps(apertures_info)

    hdul = fits.HDUList([prim_hdu, spec_hdu, var_hdu])
    return hdul, wave


def main():
    """
    Generate synthetic spectra.

    Returns
    -------
    None.

    """
    warnings.simplefilter('ignore', category=AstropyWarning)

    if not os.path.exists(TEST_DATA_PATH):
        os.makedirs(TEST_DATA_PATH)

    wave_range = [4500, 9000]
    obj_id = 0
    type_counters = {}

    spec_files = []

    for t_dict in get_sdss_spectral_templates(outdir=TEST_DATA_PATH):
        # Get a random redshift from the template redshifts range
        obj_type = t_dict['type'].lower()

        if obj_type not in type_counters:
            type_counters[obj_type] = 0
        else:
            type_counters[obj_type] = type_counters[obj_type] + 1

        z = 0

        if obj_type == 'galaxy':
            while z < 0.2:
                z = np.random.rand() * 1
        elif obj_type == 'qso':
            while z < 1.0:
                z = np.random.rand() * 2

        hdul, wave = fake_spectrum_fits(
            f'RR{obj_id:04}',
            wave_range,
            get_sdss_template_data(t_dict['file']),
            z
        )

        outname = f"specex_sdss_{obj_type}"
        try:
            obj_sub_type = t_dict['sub_type'].lower()
        except KeyError:
            pass
        else:
            outname += f'_{obj_sub_type}'

        out_file_path = os.path.join(
            TEST_DATA_PATH,
            f"{outname}_{type_counters[obj_type]:02d}.fits"
        )
        hdul.writeto(out_file_path, overwrite=True)

        spec_files.append(out_file_path)
        obj_id += 1
    return spec_files


if __name__ == '__main__':
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import specex
sys.path.insert(0, os.path.abspath('../src'))

# -- Project information -----------------------------------------------------

project = 'python-specex'
copyright = "2022-2024, Maurizio D'Addona <mauritiusdadd@gmail.com>"
author = "Maurizio D'Addona"

# The full version, including alpha/beta/rc tags
release = specex.__version__

github_url = "https://github.com/mauritiusdadd/python-specex"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx_copybutton',
]

issuetracker = 'github'
issuetracker_project = 'mauritiusdadd/python-specex'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    'css/custom.css',
]

html_context = {
    "display_github": True, # Integrate GitHub
    "github_user": "mauritiusdadd", # Username
    "github_repo": "python-specex", # Repo name
    "github_version": "main", # Version
    "conf_py_path": "/docs/", # Path in the checkout to the docs root
}

# EPUB options
epub_show_urls = 'footnote'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

This module provides utility functions used by other specex modules.

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
import os
import sys
import time
import tarfile
import logging
import platform
import subprocess
from typing import Optional, Union, NoReturn
from urllib import request

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib import patches
import matplotlib.patheffects as PathEffects

from scipy.signal import savgol_filter
from scipy.ndimage import rotate
from astropy.io import fits
from astropy.visualization import ZScaleInterval
from astropy import wcs as apwcs
from astropy import units as apu
from astropy import coordinates
from astropy import constants
from astropy.table import Table

from specex.lines import get_lines

try:
    from regions import Regions
except Exception:
    HAS_REGION = False
else:
    HAS_REGION = True

try:
    import reproject
except Exception:
    HAS_REPROJECT = False
    def do_reprojectdef(img_hdu, cube_hdu):
        raise NotImplementedError()
else:
    HAS_REPROJECT = True

    def do_reproject(img_hdu, sp_cube):
        return reproject.reproject_exact(
            img_hdu,
            sp_cube.getSpecWCS().celestial,
            sp_cube.spec_hdu.data.shape[1:],
            parallel=True,
        )

_SDSS_SPECTRAL_TEMPLATES_PACKAGE = "http://classic.sdss.org/dr5/algorithms/spectemplates/spectemplatesDR2.tar.gz"


def static_vars(**kwargs):
    """
    Add attributes to a function.

    Parameters
    ----------
    **kwargs : TYPE
        Keyword arguments will be added as attributes.

    Returns
    -------
    func
        The decorator function.

    """
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate


class ScaleBar:
    """A simple scalebar."""

    def __init__(self, ax, x0=0.025, y0=0.025, length=1, size=0.025,
                 orientation='hor', n_subunits=2, text=None, fontsize=8,
                 units='arcsec', scale_factor=1):

        self.ax = ax
        self.units = units
        self.start_pos = (x0, y0)
        self.scale_factor = scale_factor
        self.size = size

        if orientation.lower().startswith('h'):
            self.end_pos = (x0 + length, y0)
        else:
            self.end_pos = (x0, y0 + length)

        self.orientation = orientation
        self.n_subunits = n_subunits
        self.length = length
        self.scale_bar_elements = []

        if text is None:
            text = f"{length:.1f}"

        length, _ = (
            ax.transData + ax.transAxes.inverted()
        ).transform([length, 0])

        self.text_handler = ax.text(
            x0 + length / 2,
            y0 + 2 * size,
            text,
            horizontalalignment='center',
            verticalalignment='bottom',
            transform=ax.transAxes,
            fontsize=fontsize
        )

        self.text_handler.set_path_effects(
            (
                PathEffects.withStroke(
                    linewidth=3, foreground='white'
                ),
            )
        )

        delta = length / n_subunits

        for i in range(n_subunits):
            if orientation.lower().startswith('h'):
                x = x0 + i*delta
                y = y0
                wid = delta
                hei = size
            else:
                x = x0
                y = y0 + i*delta
                wid = size
                hei = delta

            if i % 2:
                facecolor = 'white'
                edgecolor = 'black'
            else:
                facecolor = 'black'
                edgecolor = 'white'

            scale_bar_element = patches.Rectangle(
                (x, y),
                wid,
                hei,
                transform=ax.transAxes,
                facecolor=facecolor,
                edgecolor=edgecolor,
                lw=1,
                ls='-',
            )

            ax.add_artist(scale_bar_element)
            self.scale_bar_elements.append(scale_bar_element)

    def update(self):
        """
        Update the ScaleBar.

        Returns
        -------
        None.

        """
        length, _ = (
            self.ax.transData + self.ax.transAxes.inverted()
        ).transform([self.length, 0])

        delta = length / self.n_subunits

        self.text_handler.set_x(self.start_pos[0] + length / 2)
        self.text_handler.set_y(self.start_pos[1] + 2 * self.size)

        for i, rect in enumerate(self.scale_bar_elements):
            if self.orientation.lower().startswith('h'):
                x = self.start_pos[0] + i*delta
                y = self.start_pos[1]
                wid = delta
                hei = self.size
            else:
                x = self.start_pos[0]
                y = self.start_pos[1] + i*delta
                wid = self.size
                hei = delta

            rect.set_x(x)
            rect.set_y(y)
            rect.set_width(wid)
            rect.set_height(hei)

    def set_wcs(self, wcs):
        """
        Set the WCS.

        Parameters
        ----------
        wcs : astropy.wcs.WCS
            The WCS to use.

        Returns
        -------
        None.

        """
        star_coords = wcs.pixel_to_world(
            *self.start_pos
        )

        end_coords = wcs.pixel_to_world(
            *self.end_pos
        )

        sep = star_coords.separation(end_coords).to(self.units)
        sep /= self.scale_factor

        self.text_handler.set_text(f"{sep:.2f}")
        self.update()


def find_prog(prog):
    """
    Find the path of a pregram in your PATH.

    Parameters
    ----------
    prog : str
        Name of the program.

    Returns
    -------
    str
        The path of the program.

    """
    cmd = "where" if platform.system() == "Windows" else "which"
    try:
        return subprocess.check_output([cmd, prog]).strip().decode()
    except subprocess.CalledProcessError:
        return None


def get_pbar(partial, total=None, wid=32, common_char='\u2588',
             upper_char='\u2584', lower_char='\u2580'):
    """
    Return a nice text/unicode progress bar showing partial and total progress.

    Parameters
    ----------
    partial : float
        Partial progress expressed as decimal value.
    total : float, optional
        Total progress expresses as decimal value.
        If it is not provided or it is None, than
        partial progress will be shown as total progress.
    wid : int , optional
        Width in charachters of the progress bar.
        The default is 32.

    Returns
    -------
    pbar : str
        A unicode progress bar.

    """
    wid -= 2
    prog = int((wid)*partial)
    if total is None:
        total_prog = prog
        common_prog = prog
    else:
        total_prog = int((wid)*total)
        common_prog = min(total_prog, prog)
    pbar_full = common_char*common_prog
    pbar_full += upper_char*(total_prog - common_prog)
    pbar_full += lower_char*(prog - common_prog)
    return (f"\u2595{{:<{wid}}}\u258F").format(pbar_full)


@static_vars(prev_iter_text_len=0)
def loop_print(text, file=sys.stdout, end=None):
    """
    Print a single line of text.

    This function can be used to print a row of text that should be
    overwritten in a next iteration.

    Remember to use loop_print.reset() before a new loop.

    Parameters
    ----------
    text : TYPE
        DESCRIPTION.
    file : TYPE, optional
        DESCRIPTION. The default is sys.stdout.

    Returns
    -------
    None.

    """
    text_len = len(text)
    delta_str_len = loop_print.prev_iter_text_len - text_len
    if delta_str_len > 0:
        text += " " * (delta_str_len)
    file.write(f"\r{text} \r")
    if end:
        file.write(end)
        loop_print.prev_iter_text_len = 0
    else:
        loop_print.prev_iter_text_len = text_len
    file.flush()


@static_vars(last_time=0)
def simple_pbar_callback(k, total, min_interval=1):
    """
    Print a progressbar with a cooldown interval.

    Parameters
    ----------
    k : float
        DESCRIPTION.
    total : float
        DESCRIPTION.
    min_interval : TYPE, optional
        Miniumum interval in seconds between two printings. The default is 1.

    Returns
    -------
    None.

    """
    if (k <= 0):
        loop_print.prev_iter_text_len = 0
        simple_pbar_callback.last_time = 0

    if (
        (k <= 1) or
        (k == total) or
        (time.perf_counter() - simple_pbar_callback.last_time) > min_interval
    ):
        simple_pbar_callback.last_time = time.perf_counter()
        pbar = get_pbar(k / total)
        loop_print(f"\r{pbar} {k / total:.2%} \r")

    if (k >= total):
        loop_print.prev_iter_text_len = 0
        simple_pbar_callback.last_time = 0


def get_sdss_spectral_templates(outdir: str, use_cached: bool = True) -> list:
    """
    Download spectral templates from SDSS.

    Parameters
    ----------
    outdir : str
        Path where to save the templates.

    use_cached : bool, optional
        If true, do not redownload the package file if it is still present.
        Default is True.

    Returns
    -------
    templates : list
        A a list of dictionaries that have the following structure:

        'file' : str
            The path of the actual template file
        'type' : str
            The type of template. Can be 'star', 'galaxy' or 'qso'
        'sub-type' : str
            The sub-type of the object
    """
    def report_pbar(blocks_count, block_size, total_size):
        downloaded_size = blocks_count * block_size
        progress = downloaded_size / total_size
        pbar = get_pbar(progress)
        report_str = f"\r{pbar} {progress: 6.2%}  "
        report_str += f"{downloaded_size}/{total_size} Bytes\r"
        sys.stderr.write(report_str)
        sys.stderr.flush()

    def same_dest_dir(outdir, member):
        member_path = os.path.join(outdir, member.name)
        abs_dest_dir = os.path.abspath(outdir)
        abs_target_dir = os.path.abspath(member_path)
        prefix = os.path.commonprefix([abs_dest_dir, abs_target_dir])
        return prefix == abs_dest_dir

    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    package_name = os.path.basename(_SDSS_SPECTRAL_TEMPLATES_PACKAGE)
    package_out_file = os.path.join(outdir, package_name)

    print(
        f"\nDownloading SDSS spectral templates package {package_name}...",
        file=sys.stderr
    )

    if not (use_cached and os.path.isfile(package_out_file)):
        package_out_file, headers = request.urlretrieve(
            _SDSS_SPECTRAL_TEMPLATES_PACKAGE,
            package_out_file,
            reporthook=report_pbar
        )

    print(
        "\nExtracting template files\n",
        file=sys.stderr
    )

    templates = []
    with tarfile.open(package_out_file, 'r') as gzfile:
        members_list = gzfile.getmembers()
        for j, member in enumerate(members_list):
            progress = j / len(members_list)
            pbar_str = get_pbar(progress)
            sys.stderr.write(f"\r{pbar_str} {progress: 6.2%}  {member.name}\r")
            sys.stdout.flush()
            if not same_dest_dir(outdir, member):
                print(
                    "Attempted Path Traversal in Tar File for member "
                    f"{member.name}",
                    file=sys.stdout
                )
                continue
            gzfile.extract(member, path=outdir, numeric_owner=False)
            t_path = os.path.abspath(os.path.join(outdir, member.name))
            t_type_id = os.path.splitext(os.path.basename(member.name))[0]
            t_type_id = int(t_type_id[-3:])

            if t_type_id <= 23:
                t_type = 'star'
            elif t_type_id <= 29:
                t_type = 'galaxy'
            else:
                t_type = 'qso'

            templates.append({
                'file': t_path,
                'type': t_type,
                'sub-type': ''
            })

    return templates


def get_sdss_template_data(sdss_template_file: str) -> dict:
    """
    Get spectrum from a SDSS spectral template file.

    Parameters
    ----------
    sdss_template_file : str
        Path of a SDSS spectral template file.

    Returns
    -------
    t_dict : dict
        A dictionary contatining the following keys and values:
            'flux' : numpy.ndarray 1D
                The flux data in arbitrary units
            'wavelengths' : numpy.ndarray 1D
                The wavelengths
    """
    try:
        t_spectrum = fits.getdata(sdss_template_file, ext=0)[0]
        t_head = fits.getheader(sdss_template_file, ext=0)
    except (FileNotFoundError, IndexError):
        return {}
    else:
        t_wcs = apwcs.WCS(t_head)

    pixel_coords_x = np.arange(0, len(t_spectrum), 1)
    pixel_coords_y = np.zeros_like(pixel_coords_x)

    t_wavelengths = t_wcs.pixel_to_world_values(pixel_coords_x, pixel_coords_y)
    t_wavelengths = 10 ** t_wavelengths[0]

    t_dict = {
        'flux': t_spectrum,
        'wavelengths': t_wavelengths
    }
    return t_dict


def parse_regionfile(regionfile, key_ra='ALPHA_J2000', key_dec='DELTA_J2000',
                     key_id='NUMBER', file_format='ds9'):
    """
    Parse a regionfile and return an asrtopy Table with sources information.

    Note that the only supported shape are 'circle', 'ellipse' and 'box',
    other shapes in the region file will be ignored. Note also that 'box'
    is treated as the bounding box of an ellipse.

    Parameters
    ----------
    regionfile : str
        Path of the regionfile.
    pixel_scale : float or None, optional
        The pixel scale in mas/pixel used to compute the dimension of the
        size of the objects. If None, height and width in the region file will
        be considered already in pixel units.
    key_ra : str, optional
        Name of the column that will contain RA of the objects.
        The default value is 'ALPHA_J2000'.
    key_dec : str, optional
        Name of the column that will contain DEC of the objects
        The default value is 'DELTA_J2000'.
    file_format : str, optional
        Format of the input regionfile.
        The default value is 'ds9'.

    Returns
    -------
    sources : astropy.table.Table
        The table containing the sources.
    skyframe : None
        Placeholder.

    """
    global HAS_REGION

    if not HAS_REGION:
        logging.error(
            "astropy regions package is needed to handle regionfiles!"
        )
        return

    myt = Table(
        names=[key_id, key_ra, key_dec, 'region'],
        units=[None, 'deg', 'deg', None],
        dtype=[str, float, float, object]
    )
    for j, reg in enumerate(Regions.read(regionfile, format=file_format)):
        try:
            reg_id = str(reg.meta['text'])
        except Exception:
            reg_id = str(j)

        try:
            center = reg.center
        except AttributeError:
            c_ra = np.mean([x.ra.to('deg').value for x in reg.vertices])
            c_dec = np.mean([x.dec.to('deg').value for x in reg.vertices])
            new_row = [reg_id, c_ra, c_dec, reg]
        else:
            new_row = [reg_id, center.ra, center.dec, reg]
        myt.add_row(new_row)

    return myt, None


def get_aspect(ax):
    """
    Get ratio between y-axis and x-axis of a matplotlib figure.

    Parameters
    ----------
    ax : matplotlib.axes._subplots.AxesSubplot
        Ther axis you want to get the axes ratio.

    Returns
    -------
    ratio : float
        The aspect ratio.

    """
    figW, figH = ax.get_figure().get_size_inches()

    # Axis size on figure
    _, _, w, h = ax.get_position().bounds

    # Ratio of display units
    disp_ratio = (figH * h) / (figW * w)

    # Ratio of data units
    data_ratio = (max(*ax.get_ylim()) - min(*ax.get_ylim()))
    data_ratio /= (max(*ax.get_xlim()) - min(*ax.get_xlim()))

    return disp_ratio / data_ratio


def get_vclip(img, vclip=0.25, nsamples=1000):
    """
    Get the clipping values to use with imshow function.

    Parameters
    ----------
    img : numpy.ndarray
        Image data.
    vclip : float, optional
        Contrast parameter. The default is 0.5.

    Returns
    -------
    vmin : float
        median - vclip*std.
    vmax : float
        median + vclip*std.

    """
    img = np.ma.masked_array(img, mask=np.isnan(img))
    zsc = ZScaleInterval(nsamples, contrast=vclip, krej=10)
    vmin, vmax = zsc.get_limits(img)
    return vmin, vmax


def get_log_img(img, vclip=0.5):
    """
    Get the image in log scale.

    Parameters
    ----------
    img : numpy.ndarray
        The image data.
    vclip : float, optional
        Contrast factor. The default is 0.5.

    Returns
    -------
    log_image : numpy.ndarray
        The logarithm of the input image.
    vclip : 2-tuple of floats
        The median +- vclip*std of the image.

    """
    log_img = np.log10(1 + img - np.nanmin(img))
    return log_img, *get_vclip(log_img)


def load_rgb_fits(fits_file, ext_r=1, ext_g=2, ext_b=3):
    """
    Load an RGB image from a FITS file.

    Parameters
    ----------
    fits_file : str
        The path of the fits file.
    ext_r : int, optional
        The index of the extension containing the red channel.
        The default is 1.
    ext_g : int, optional
        The index of the extension containing the green channel.
        The default is 2.
    ext_b : int, optional
        The index of the extension containing the blue channel.
        The default is 3.

    Returns
    -------
    dict
        The dictionary contains the following key: value pairs:
            'data': 3D numpy.ndarray
                The actual image data.
            'wcs': astropy.wcs.WCS
                The WCS of the image.
    """
    try:
        rgb_image = np.array((
            fits.getdata(fits_file, ext=ext_r),
            fits.getdata(fits_file, ext=ext_g),
            fits.getdata(fits_file, ext=ext_b),
        )).transpose(1, 2, 0)
    except IndexError:
        return None

    rgb_image -= np.nanmin(rgb_image)
    rgb_image = rgb_image / np.nanmax(rgb_image)

    rgb_wcs = apwcs.WCS(fits.getheader(fits_file, ext=1))

    return {'data': rgb_image, 'wcs': rgb_wcs, 'type': 'rgb'}


def log_rebin(wave, spec_list, oversample=1, flux_conserving=False):
    """
    Logarithmically rebin a spectrum conserving the flux.

    Parameters
    ----------
    wave : list or numpy.ndarray
        The wavelengths of the originaò dispersion grating.
    spec_list : list
        A list of numpy.ndarrays corrsponding toe the fulx or other quantities
        to be rebinned with the dispersion grating.
    oversample : int, optional
        Whether to oversample the new dispersion grating. The default is 1.
    flux_conserving : bool, optional
        If False conserve the flux density instead of the flux.
        The default is False.

    Raises
    ------
    ValueError
        If wave is not monotonically increasing or if it does not have the
        same dimension of the fluxes.

    Returns
    -------
    rebinned_spec_list : list
        A list containing the rebinned input fluxes.
    log_wave : numpy.ndarray
        The log-reminned disperions grating.
    velscale : float
        The velocity scale of the new dispersion grating.

    """
    wave = np.asarray(wave, dtype=float)
    spec_list = [
        np.asarray(x, dtype=float) for x in spec_list
    ]

    if np.any(np.diff(wave) <= 0):
        raise ValueError("'wave' must be monotonically increasing")

    if (
            (len(wave.shape) != 1) or
            np.any([wave.shape[0] != x.shape[0] for x in spec_list])
    ):
        raise ValueError(
            "'wave' must be an array with the same lenght of the input spectra"
        )

    if wave.shape[0] == 2:
        dlam = np.diff(wave) / (wave.shape[0] - 1)
        lim = wave + [-0.5, 0.5]*dlam
        borders = np.linspace(*lim, wave.shape[0] + 1)
    else:
        lim = 1.5*wave[[0, -1]] - 0.5*wave[[1, -2]]
        borders = np.hstack([lim[0], (wave[1:] + wave[:-1])/2, wave[1]])
        dlam = np.diff(borders)

    ln_lim = np.log(lim)
    c_km_h = constants.c.to(apu.km / apu.h).value

    m = int(wave.shape[0] * oversample)
    velscale = c_km_h * np.diff(ln_lim) / m
    velscale = velscale.item()

    new_borders = np.exp(ln_lim[0] + velscale / c_km_h * np.arange(m + 1))

    if wave.shape[0] == 2:
        k = ((new_borders - lim[0])/dlam)
        k = k.clip(0, wave.shape[0] - 1).astype(int)
    else:
        k = (np.searchsorted(borders, new_borders) - 1)
        k = k.clip(0, wave.shape[0] - 1).astype(int)

    # Do analytic integral of step function
    rebinned_spec_list = []
    for spec in spec_list:
        spec_rebin = np.add.reduceat((spec.T*dlam).T, k)[:-1]
        # fix for design flaw of reduceat()
        spec_rebin.T[...] *= np.diff(k) > 0
        # Add to 1st dimension
        spec_rebin.T[...] += np.diff(((new_borders - borders[k])) * spec[k].T)

        if not flux_conserving:
            # Divide 1st dimension
            spec_rebin.T[...] /= np.diff(new_borders)

        rebinned_spec_list.append(spec_rebin)

    # Output np.log(wavelength): natural log of geometric mean
    ln_lam = 0.5*np.log(new_borders[1:]*new_borders[:-1])

    return rebinned_spec_list, ln_lam, velscale


def plot_scandata(target, scandata):
    """
    Plot debugging information about zchi2.

    Parameters
    ----------
    target : redrock.Target
        The target object.
    scandata : dict
        The full scandata as returned by redrock.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        The figure of the plot.
    ax : TYPE
        The main axis of the plot.

    """
    try:
        obj_data = scandata[target.id]
    except KeyError:
        return

    template_types = list(obj_data.keys())

    fig, ax = plt.subplots(1, 1, figsize=(10, 5))

    for template_type in template_types:
        ax.plot(
            obj_data[template_type]['redshifts'],
            obj_data[template_type]['zchi2'],
            label=f"{template_type}"
        )
    ax.set_yscale('log')
    ax.set_xscale('log')
    ax.set_xlim(left=0.1)
    ax.set_title(f'{target.id}')
    ax.set_xlabel("redshift")
    ax.set_ylabel("Chi2")
    ax.legend()
    plt.tight_layout()
    return fig, ax


def get_ellipse_skypoints(center: coordinates.SkyCoord,
                          a: apu.quantity.Quantity,
                          b: apu.quantity.Quantity,
                          angle: apu.quantity.Quantity = 0*apu.deg,
                          n_points: int = 20) -> list:
    """
    Get points of an ellips on the skyplane.

    Parameters
    ----------
    center : astropy.coordinates.SkyCoord
        DESCRIPTION.
    a : apu.quantity.Quantity
        Angular size of the semi-major axis.
    b : apu.quantity.Quantity
        Angular size of the semi-mino axis.
    angle : apu.quantity.Quantity, optional
        Rotation angle of the semi-major axis respect to the RA axis.
        The default is 0 deg.
    n_points : int, optional
        Number of points to return. The default is 25.

    Returns
    -------
    ellipse_points : list of coordinates.SkyCoord
        List of SkyCoord corresponding to the points of the ellipse.

    Example
    -------
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from astropy.io import fits
    >>> from astropy.wcs import WCS
    >>> from astropy.wcs.utils import wcs_to_celestial_frame
    >>> from astropy.coordinates import SkyCoord
    >>> from astropy import units as apu
    >>> from astropy.utils.data import get_pkg_data_filename
    >>> from specex.utils import get_ellipse_skypoints
    >>> fn = get_pkg_data_filename('galactic_center/gc_msx_e.fits')
    >>> image = fits.getdata(fn, ext=0)
    >>> w = WCS(fits.getheader(fn, ext=0))
    >>> ellipse_center = w.pixel_to_world(50, 60)
    >>> a = 10 * apu.arcmin
    >>> b = 5 * apu.arcmin
    >>> angle = 45 * apu.deg
    >>> world_points = get_ellipse_skypoints(
    ...     center=ellipse_center,
    ...     a=a, b=b,
    ...     angle=angle
    ... )
    >>> pp_val = np.array([
    ...     [x.l.value, x.b.value] for x in world_points
    ... ])
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(projection=w)
    >>> ax.imshow(
    ...     image,
    ...     vmin=-2.e-5,
    ...     vmax=2.e-4,
    ...     origin='lower',
    ...     cmap='plasma'
    ... )
    >>> ax.plot(
    ...     pp_val[..., 0], pp_val[..., 1],
    ...     color='#31cc02',
    ...     lw=2,
    ...     ls='--',
    ...     transform=ax.get_transform(wcs_to_celestial_frame(w))
    ... )
    >>> ax.set_aspect(1)
    >>> plt.show()
    >>> plt.close(fig)
    """
    ellipse_points = []

    # Check if a is actually greater than b, otherwise swap them
    if a < b:
        _tmp = a
        a = b
        b = _tmp

    for theta in np.linspace(0, 2*np.pi, n_points, endpoint=True):
        total_angle = -theta + angle.to(apu.rad).value
        radius = a*b / np.sqrt(
            (a*np.cos(total_angle))**2 + (b*np.sin(total_angle))**2
        )
        new_point = center.directional_offset_by(
            position_angle=apu.Quantity(theta, apu.rad),
            separation=radius
        )

        ellipse_points.append(new_point)
    return ellipse_points


def plot_masked_regions(ax, wavelengths, nan_mask, wave_range=None,
                        label_min_wid=100):
    """
    Plot a mask over a region containing missing data.

    Parameters
    ----------
    ax : TYPE
        DESCRIPTION.
    wavelengths : TYPE
        DESCRIPTION.
    nan_mask : TYPE
        DESCRIPTION.
    wave_range : TYPE, optional
        DESCRIPTION. The default is None.
    label_min_wid : intm optional
        The default value is 100.

    Returns
    -------
    region_patches : dict
        DESCRIPTION.

    """
    region_patches = {
        'background': [],
        'area': [],
        'text': []
    }

    if nan_mask is None:
        return region_patches

    if wave_range is None:
        w_min = np.nanmin(wavelengths)
        w_max = np.nanmax(wavelengths)
    else:
        w_min = np.min(wave_range)
        w_max = np.max(wave_range)

    lam_mask = np.array([
        (wavelengths[m_start], wavelengths[m_end])
        for m_start, m_end in get_mask_intervals(nan_mask)
    ])

    w_min = np.nanmin(wavelengths)
    w_max = np.nanmax(wavelengths)

    for lam_inter in lam_mask:
        rect = patches.Rectangle(
            (lam_inter[0], 0),
            lam_inter[1] - lam_inter[0], 1,
            transform=ax.get_xaxis_transform(),
            linewidth=1,
            fill=True,
            edgecolor='black',
            facecolor='white',
            zorder=10
        )
        ax.add_patch(rect)
        region_patches['background'].append(rect)

        rect = patches.Rectangle(
            (lam_inter[0], 0),
            lam_inter[1] - lam_inter[0], 1,
            transform=ax.get_xaxis_transform(),
            linewidth=1,
            fill=True,
            edgecolor='black',
            facecolor='white',
            hatch='///',
            zorder=11
        )
        ax.add_patch(rect)
        region_patches['area'].append(rect)
        if (
            (lam_inter[1] > w_min + label_min_wid) and
            (lam_inter[0] < w_max - label_min_wid)
        ):
            text = ax.text(
                (lam_inter[0] + lam_inter[1]) / 2, 0.5,
                "MISSING DATA",
                transform=ax.get_xaxis_transform(),
                va='center',
                ha='center',
                rotation=90,
                bbox={
                    'facecolor': 'white',
                    'edgecolor': 'black',
                    'boxstyle': 'round,pad=0.5',
                },
                zorder=12,
                clip_on=True
            )
            region_patches['text'].append(text)
    return region_patches


def plot_lines(ax, wave_range, lw=0.7, lines_z=0, label_size='medium',
               colors=['green', 'red', 'yellow'], label_y=0.02,
               alpha=[0.5, 0.75, 0.9], ls=['--', '--', '--'],
               label_rotation=90,):
    """
    Plot emission and/or absorption lines.

    Parameters
    ----------
    ax : matplotlib Axes
        The Axes where to draw the lines.
    wave_range : liat or numpy.ndarray
        wavelength range.
    lw : float, optional
        width of the plotted lines. The default is 0.7.
    lines_z : float, optional
        The redshift. The default is 0.
    label_size : float or str, optional
        The size of the labels. The deafult value is 'medium'.
    colors : list of 3 colors, optional
        The color used to draw the different type of lines. The first color is
        assinged to absorption lines, the second one to emission lines and the
        third one to lines that can be both absoprtion and emission.
        The default value is ['green', 'red', 'yellow'].
    label_y : float, optional
        The position of the labels in axis-coordinates.
        The default value is 0.02
    alpha : list of 3 floats, optional
        The value of the alpha transparence used to draw the lines. The order
        is the same of the colors array. The default value is [0.5, 0.75, 0.9].
    ls : list of 3 line styles, optional
        Styles used to plot lines. The order is the same of the colors array.
        The default value is ['--', '--', '--']
    label_rotation : float, optional
        The rotation in degrees of the labels. The default value is 90

    Returns
    -------
    None.

    """
    w_min = np.min(wave_range)
    w_max = np.max(wave_range)

    absorption_lines = get_lines(
        line_type='A', wrange=[w_min, w_max], z=lines_z
    )
    for line_lam, line_name, line_type in absorption_lines:
        ax.axvline(
            line_lam, color=colors[0], ls='--', lw=lw, alpha=alpha[0],
            label='absorption lines'
        )
        ax.text(
            line_lam, label_y, line_name, rotation=label_rotation,
            transform=ax.get_xaxis_transform(),
            zorder=99,
            fontsize=label_size
        )

    # Plotting emission lines
    emission_lines = get_lines(
        line_type='E', wrange=[w_min, w_max], z=lines_z
    )
    for line_lam, line_name, line_type in emission_lines:
        ax.axvline(
            line_lam, color=colors[1], ls='--', lw=lw, alpha=alpha[1],
            label='emission lines',
            zorder=2
        )
        ax.text(
            line_lam, label_y, line_name, rotation=label_rotation,
            transform=ax.get_xaxis_transform(),
            zorder=99,
            fontsize=label_size
        )

    # Plotting emission/absorption lines
    emission_lines = get_lines(
        line_type='AE', wrange=[w_min, w_max], z=lines_z
    )
    for line_lam, line_name, line_type in emission_lines:
        ax.axvline(
            line_lam, color=colors[2], ls='--', lw=lw, alpha=alpha[2],
            label='emission/absorption lines',
            zorder=3
        )
        ax.text(
            line_lam, label_y, line_name, rotation=label_rotation,
            transform=ax.get_xaxis_transform(),
            zorder=99,
            fontsize=label_size
        )


def plot_spectrum(wavelengths, flux, variance=None, nan_mask=None,
                  restframe=False, cutout=None, cutout_vmin=None,
                  cutout_vmax=None, cutout_wcs=None, redshift=None,
                  smoothing=None, wavelengt_units=None, flux_units=None,
                  extra_info={}, extraction_info={}, wave_range=None,
                  show_legend=True, axs=None):
    """
    Plot a spectrum.

    Parameters
    ----------
    wavelengths : numpy.ndarray 1D
        An array containing the wavelengths.
    flux : numpy.ndarray 1D
        An array containing the fluxes corresponding to the wavelengths.
    variance : numpy.ndarray 1D, optional
        An array containing the variances corresponding to the wavelengths.
        If it is None, then the variance is not plotted. The default is None.
    nan_mask : numpy.ndarray 1D of ndtype=bool, optional
        An array of dtype=bool that contains eventual invalid fluxes that need
        to be masked out (ie. nan_mask[j] = True means that wavelengths[j],
        flux[j] and variance[j] are masked). If it is None, then no mask is
        applyed. The default is None.
    restframe : bool, optional
        If True, then the spectrum is plotted in the observer restframe
        (ie. the spectrum is de-redshifted before plotting). In order to use
        this option, a valid redshift must be specified.
        The default is False.
    cutout : numpy.ndarray 2D or 3D, optional
        A grayscale or RGB image to be shown alongside the spectrum.
        If None, no image is shown. The default is None.
    cutout_wcs : astropy.wcs.WCS, optional
        An optional WCS for the cutout. The default is None.
    cutout_vmin : float, optional
        The value to be interpreted as black in the cutout image.
        If it is None, the value is determined automatically.
        The default is None.
    cutout_vmax : float, optional
        The value to be interpreted as white in the cutout image.
        If it is None, the value is determined automatically.
        The default is None.
    redshift : float, optional
        The redshift of the spectrum. If None then no emission/absorption
        line is plotted and restframe option is ignored. The default is None.
    smoothing : int or None, optional
        If an integer greater than zero is passed to this parameter, then the
        value is used as the window size of scipy.filter.savgol_filter used to
        smooth the flux of the spectum. If this value is 0 or None then no
        smoothing is performed. The default is None.
    wavelengt_units : str, optional
        The units of the wavelengths. The default is None.
    flux_units : str, optional
        The units of the fluxes. The default is None.
    extra_info : dict of {str: str, ...}, optional
        A dictionary containing extra information to be shown in the plot.
        Both keys and values of the dictionary must be strings. This dict is
        rendered as a table of two columns filled with the keys and the values.
        The default is {}.
    extraction_info: dict, optionale
        This dictionary must contain extraction information from specex. If not
        empty or None, extraction information are used to plot the apertures
        used by specex over the cutout (if provided).
        The default is {}.
    wave_range: list, optional
        Range of wavelengths to plot. The default value is None.
    show_legend: bool, optional
        Show the legend if true. Default value it True.
    axs: list of matplotlib.axes._subplots.AxesSubplots, optional
        List of axes to use for plotting. If none a new figure with the
        appropriate axes is created. The default value is None.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure containing the plots.
    axs: list of matplotlib.axes._subplots.AxesSubplots
        A list of the axes subplots in the figure.
    """
    if nan_mask is not None:
        if variance is not None:
            var_max = np.nanmax(variance[~nan_mask])
        else:
            var_max = 1
    else:
        if variance is not None:
            var_max = np.nanmax(variance)
        else:
            var_max = 1

    w_min = 1.0e5
    w_max = 0.0

    if restframe and redshift is not None:
        wavelengths = wavelengths / (1 + redshift)
        lines_z = 0
    else:
        wavelengths = wavelengths
        lines_z = redshift

    if wave_range is None:
        w_min = np.nanmin(wavelengths)
        w_max = np.nanmax(wavelengths)
    else:
        w_min = np.min(wave_range)
        w_max = np.max(wave_range)

    if wavelengt_units:
        try:
            x_unit = apu.Unit(wavelengt_units).to_string(
                'latex', fraction='inline'
            )
        except Exception:
            x_label = f'Wavelength [{wavelengt_units}]'
        else:
            x_label = f'Wavelength [{x_unit}]'
    else:
        x_label = 'Wavelength'

    if flux_units:
        try:
            y_unit = apu.Unit(flux_units).to_string(
                'latex', fraction='inline'
            )
        except Exception:
            y_label = f'Flux [{flux_units}]'
        else:
            y_label = f'Flux [{y_unit}]'
    else:
        y_label = 'Flux'

    if axs is None:
        fig = plt.figure(figsize=(15, 5))

        # Make a grid of plots
        gs = GridSpec(6, 6, figure=fig, hspace=0.1)
    else:
        fig = axs[0].figure

    # If variance data are present, then make two plots on the left of the
    # figure. The top one is for the spectrum and the bottom one is for the
    # variance. Otherwise just make a bigger plot on the left only for the
    # spectrum.
    if variance is not None:
        if axs is None:
            ax0 = fig.add_subplot(gs[:4, :-1])
            ax3 = fig.add_subplot(gs[4:, :-1], sharex=ax0)
        else:
            ax0 = axs[0]
            ax3 = axs[1]

        ax3.plot(
            wavelengths, variance,
            ls='-',
            lw=0.5,
            alpha=0.75,
            color='black',
            label='variance',
            zorder=0
        )
        ax3.set_xlabel(x_label)
        ax3.set_ylabel('Variance')

        ax3.set_xlim(w_min, w_max)
        ax3.set_ylim(1, var_max)
        ax3.set_yscale('log')
        ax0.label_outer()
    else:
        if axs is None:
            ax0 = fig.add_subplot(gs[:, :-1])
            ax3 = None
        else:
            ax0 = axs[0]
            ax3 = axs[1]
        ax0.set_xlabel(x_label)

    ax0.set_ylabel(y_label)
    ax0.set_xlim(w_min, w_max)

    # Plot a cutout
    if cutout is not None:
        if axs is None:
            ax1 = fig.add_subplot(gs[:3, -1], projection=cutout_wcs)
            ax2 = fig.add_subplot(gs[3:, -1])
        else:
            ax1 = axs[2]
            ax2 = axs[3]

        ax1.axis('off')
        ax1.imshow(
            cutout,
            origin='lower',
            aspect='equal',
            vmin=cutout_vmin,
            vmax=cutout_vmax,
            zorder=0
        )
        ax1.set_aspect(1)

        # Check if there are info about the specex extraction
        try:
            ext_mode = extraction_info['mode']
            ext_apertures = extraction_info['apertures']
            e_ra = extraction_info['aperture_ra']
            e_dec = extraction_info['aperture_dec']
            e_frame = extraction_info['frame']
        except (TypeError, KeyError):
            # No extraction info present, just ignore
            pass
        else:
            # If there are extraction info, read the information
            e_wid, e_hei, e_ang = ext_apertures

            e_cc = coordinates.SkyCoord(
                e_ra, e_dec,
                unit=('deg', 'deg'),
                frame=e_frame
            )

            # and then draw extraction apertures
            if ext_mode.lower() in [
                    'kron_ellipse', 'kron_circular', 'circular_aperture'
            ]:
                e_world_points = get_ellipse_skypoints(
                    e_cc,
                    a=0.5*e_hei,
                    b=0.5*e_wid,
                    angle=e_ang
                )

                e_world_points_values = np.array([
                    [x.ra.value, x.dec.value]
                    for x in e_world_points
                ])

                ax1.plot(
                    e_world_points_values[..., 0],
                    e_world_points_values[..., 1],
                    color='black',
                    ls='-',
                    lw=1,
                    alpha=0.7,
                    zorder=1,
                    transform=ax1.get_transform(e_frame)
                )
                ax1.plot(
                    e_world_points_values[..., 0],
                    e_world_points_values[..., 1],
                    color='cyan',
                    ls='--',
                    lw=1,
                    alpha=0.7,
                    zorder=2,
                    transform=ax1.get_transform(e_frame)
                )

        scbar = ScaleBar(
            ax1,
            length=cutout.shape[1]/2
        )
        scbar.set_wcs(cutout_wcs)
        scbar.update()
    else:
        if axs is None:
            ax1 = None
            ax2 = fig.add_subplot(gs[:, -1])
        else:
            ax1 = axs[2]
            ax2 = axs[3]

    ax0.set_aspect('auto')

    # Plot only original spectrum or also a smoothed version
    if not smoothing:
        ax0.plot(
            wavelengths, flux,
            ls='-',
            lw=0.5,
            alpha=1,
            color='black',
            label='spectrum',
            zorder=0
        )
    else:
        window_size = 4*smoothing + 1
        smoothed_flux = savgol_filter(flux, window_size, 3)
        ax0.plot(
            wavelengths, flux,
            ls='-',
            lw=1,
            alpha=0.35,
            color='gray',
            label='original spectrum',
            zorder=0
        )
        ax0.plot(
            wavelengths, smoothed_flux,
            ls='-',
            lw=0.4,
            alpha=1.0,
            color='#03488c',
            label='smoothed spectrum',
            zorder=1
        )

    if redshift is not None:
        plot_lines(ax0, [w_min, w_max], lines_z=lines_z)

    # Draw missing data or invalid data regions
    _ = plot_masked_regions(ax0, wavelengths, nan_mask, wave_range)

    if show_legend:
        handles, labels = ax0.get_legend_handles_labels()
        newLabels, newHandles = [], []
        for handle, label in zip(handles, labels):
            if label not in newLabels:
                newLabels.append(label)
                newHandles.append(handle)

        _ = ax0.legend(
            newHandles, newLabels,
            loc='upper center',
            fancybox=True,
            shadow=False,
            bbox_to_anchor=(0.5, 1.15),
            ncol=len(newHandles)
        )

    if ax2 is not None:
        cell_text = [
            [f'{key}', f"{val}"] for key, val in extra_info.items()
        ]

        ax2.axis("off")
        if cell_text:
            tbl = ax2.table(
                cellText=cell_text,
                colWidths=[0.4, 0.6],
                loc='upper center'
            )
            tbl.scale(1, 1.5)

    return fig, [ax0, ax1, ax2, ax3]


def plot_zfit_check(target, zbest, plot_template=None, restframe=False,
                    wavelengt_units='Angstrom', flux_units=''):
    """
    Plot the check images for the fitted targets.

    This function will plot the spectra of the target object along with the
    spectra of the best matching tamplate and some other info.

    Parameters
    ----------
    target : redrock.targets.Target object
        A targets used in redshift estimation process.
    zfit : astropy.table.Table
        A table containing the reshift and other info of the input targets.
    plot_template : list of redrock.templates or None, optional
        If not None, plot the best matching tamplate.
    rest_frame : bool, optional
        Whether to plot the spectrum at restrframe.
    wave_units : str, optional
        The units of the wavelength grid. The default value id 'Angstrom'
    flux_units : str, optional
        The units of the spectum. The default value is ''.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The Figure object containing the plot.
    axs : 2-tuple of matplotlib.axes._subplots.AxesSubplot and/or None
        List of AxesSubplot. If no cutout was passed in input, the list will
        contain only the axes of the plot of the spectrum, otherwise two axes
        will be in the list: the first axes containing the plot of the spectrum
        and the second axes containing the cutout of the object.

    """
    flux_units = flux_units.replace('**', '^')

    t_best_data = zbest[zbest['SPECID'] == target.spec_id][0]

    info_dict = {
        'ID': f"{target.spec_id}",
        'Z': f"z: {t_best_data['Z']:.4f} ± {t_best_data['ZERR']:.2e}\n",
        'Template': f"{t_best_data['SPECTYPE']} {t_best_data['SUBTYPE']}",
        'SNR': f"{t_best_data['SN']:.2f}\n",
        'SNR (EM)': f"{t_best_data['SN_EMISS']:.2f}\n",
        'ZWARN': f"{t_best_data['ZWARN']}"
    }

    lam = target.spectra[0].wave.copy()

    fig, axs = plot_spectrum(
        lam,
        target.spectra[0].flux,
        cutout=None,
        redshift=t_best_data['Z'],
        restframe=restframe,
        wavelengt_units=wavelengt_units,
        flux_units=flux_units,
        extra_info=info_dict
    )

    best_template = None
    if plot_template:
        for t in plot_template:
            if (
                t.template_type == t_best_data['SPECTYPE'] and
                t.sub_type == t_best_data['SUBTYPE']
            ):
                best_template = t
                break

        if best_template:
            try:
                coeffs = t_best_data['COEFF'][:best_template.nbasis]
                template_flux = best_template.eval(
                    coeffs,
                    lam,
                    0 if restframe else t_best_data['Z'],
                )

                axs[0].plot(
                    lam, template_flux,
                    ls='-',
                    lw=1,
                    alpha=0.7,
                    c='red',
                    label=f'best template [{best_template.full_type}]'
                )
            except AssertionError:
                print(
                    f"Template warning for object {target.spec_id}\n"
                    f"  nbasis: {best_template.nbasis}\n"
                    f"  coeffs: {len(coeffs)}",
                    file=sys.stderr
                )

    return fig, axs


def get_mask_intervals(mask):
    """
    Get intervals where mask is True.

    Parameters
    ----------
    mask : numpy.ndarry
        The mask array.

    Returns
    -------
    regions : list of 2-tuples
        List of intervals.

    Example
    -------
    >>> mask = (0, 0, 0, 0 ,0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1 ,0 ,0)
    >>> get_mask_intervals(mask)

    [(5, 8), (10, 11), (14, 19)]
    """
    regions = []
    r_start = -1
    r_end = 0
    in_region = False
    for i, val in enumerate(mask):
        if val and not in_region:
            r_start = i
            in_region = True
        if in_region and not val:
            r_end = i-1
            in_region = False
            regions.append((r_start, r_end))
    return regions


def stack(
    data: np.ndarray,
    wave_mask: Optional[np.ndarray] = None,
    average: bool = False,
    quite: bool = False
) -> np.ndarray:
    """
    Stack the spectral cube along wavelength axis.

    Parameters
    ----------
    data : numpy.ndarray
        The spectral datacube.
    wave_mask : np.ndarray, optional
        1D wavelength mask. Wavelength corresponding to a False will not
        be used in the stacking. The default is None.
    quite : bool, optional
        Whether to avoid printing a progress bar or not.
        The default value is False.

    Returns
    -------
    new_data : numpy.ndarray
        The stacked datacube.

    """
    img_height, img_width = data.shape[1], data.shape[2]
    new_data: np.ndarray = np.zeros((img_height, img_width), dtype=float)
    new_count: np.ndarray = np.zeros((img_height, img_width), dtype=int)
    dat: np.ndarray
    for k in range(len(data)):
        dat = data[k]
        if (not quite) and (k % 10 == 0):
            progress = (k + 1) / len(data)
            sys.stdout.write(
                f"\rstacking cube: {get_pbar(progress)} {progress:.2%}\r"
            )
            sys.stdout.flush()

        if (wave_mask is None) or (wave_mask[k].any()):
            valid_mask = np.isfinite(dat)
            dat[~valid_mask] = 0.0
            new_count[valid_mask] += 1
            new_data += dat.astype(float)

    stacked = new_data / new_count.astype(float)
    if not average:
        stacked *= len(data)

    stacked[new_count == 0] = np.nan
    return  stacked


def nannmad(x, scale=1.48206, axis=None):
    """
    Compute the MAD of an array.

    Compute the Median Absolute Deviation of an array ignoring NaNs.

    Parameters
    ----------
    x : np.ndarray
        The input array.
    scale : float, optional
        A costant scale factor that depends on the distributuion.
        See https://en.wikipedia.org/wiki/Median_absolute_deviation.
        The default is 1.4826.
    axis : int or None
        The axis along which to compute the MAD.
        The default is None.

    Returns
    -------
    nmad
        The NMAD value.

    """
    x = np.ma.array(x, mask=np.isnan(x))
    x_bar = np.ma.median(x, axis=axis)
    mad = np.ma.median(np.ma.abs(x - x_bar), axis=axis)
    return scale*mad


def get_spectrum_snr(flux: np.ndarray,
                     var: Optional[np.ndarray] = None,
                     smoothing_window: Optional[int] = 51,
                     smoothing_order: Optional[int] = 11):
    """
    Compute the SRN of a spectrum.

    Parameters
    ----------
    flux : numpy.ndarray
        The spectrum itself.
    var : numpy.ndarray, optional
        The variance of the spectrum itself.
        The default value is None.
    smoothing_window : int, optional
        Parameter to be passed to the smoothing function.
        The default is 51.
    smoothing_order : int, optional
        Parameter to be passed to the smoothing function.
        The default is 11.

    Returns
    -------
    sn_spec : float
        The SNR of the spectrum.

    """
    # DER-like SNR but with a true smoothing
    # https://stdatu.stsci.edu/vodocs/der_snr.pdf
    # Smoothing the spectrum to get a crude approximation of the continuum

    if np.isnan(flux).all():
        return np.nan
    else:
        flux = np.ma.array(flux.copy(), mask=np.isnan(flux))

    if var is not None:
        var = np.ma.array(var.copy(), mask=np.isnan(var))
    else:
        var = 1.0

    smoothed_spec = savgol_filter(flux, smoothing_window, smoothing_order)
    smoothed_spec = np.ma.array(smoothed_spec, mask=np.isnan(smoothed_spec))

    # Subtract the smoothed spectrum to the spectrum itself to get a
    # crude estimation of the noise
    noise_spec = flux - smoothed_spec

    # Get the median value of the spectrum, weighted by the variance
    obj_mean_spec = np.ma.sum(smoothed_spec / var) / np.ma.sum(1 / var)

    # Get the mean Signal to Noise ratio
    sn_spec = obj_mean_spec / nannmad(noise_spec)

    if np.isinf(sn_spec):
        sn_spec = 99
    elif np.isnan(sn_spec):
        sn_spec = 0

    return sn_spec


def get_spectrum_snr_emission(flux, var=None, bin_size=150):
    """
    Compute the SRN of a spectrum considering emission lines only.

    Parameters
    ----------
    flux : numpy.ndarray
        The spectrum itself.
    bin_size : int, optional
        Bin size to search for emission lines.
        The default is 50.

    Returns
    -------
    sn_spec : float
        The SNR of the spectrum.

    """
    # Inspired by https://www.aanda.org/articles/aa/pdf/2012/03/aa17774-11.pdf

    # Just ignore negative fluxes!
    flux = flux.copy()
    flux[flux < 0] = 0

    # If we have the variace, we can use it to weight the flux
    if var is not None:
        var = var.copy()
        flux = flux / var

    optimal_width = flux.shape[0] - flux.shape[0] % bin_size
    flux = flux[:optimal_width]

    if np.isnan(flux).all():
        return np.nan
    else:
        flux = np.ma.array(flux, mask=np.isnan(flux))

    if flux.mask.all():
        return np.nan

    # Rebin sub_spec to search for emission features
    sub_spec = flux.reshape(flux.shape[0] // bin_size, bin_size)

    # For each bin we compute the maximum and the median of each bin and
    # get their difference. This is now our "signal": if there is an
    # emission line, the maximum value is greater that the median and this
    # difference will be greater than one
    sub_diff = np.ma.max(sub_spec, axis=1) - np.ma.median(sub_spec, axis=1)

    s_em = sub_diff / 3.0*np.ma.median(sub_diff) - 1
    noise_em = nannmad(sub_diff)

    sn_spec = np.ma.max(s_em / noise_em)

    if np.isinf(sn_spec):
        sn_spec = 99
    elif np.isnan(sn_spec):
        sn_spec = 0

    return sn_spec


def get_pc_transform_params(wcs_object, inverse=False, ftol=1e-6):
    cel_w = wcs_object.celestial
    pcm = cel_w.wcs.get_pc()
    if inverse:
        pcm = np.linalg.inv(pcm)
    cdelt = cel_w.wcs.get_cdelt()

    sx = cdelt[0] * np.sign(pcm[0, 0]) * np.sqrt(pcm[0, 0]**2 + pcm[0, 1]**2)
    sy = cdelt[1] * np.sign(pcm[1, 0]) * np.sqrt(pcm[1, 0]**2 + pcm[1, 1]**2)
    rot = np.arctan2(-pcm[0, 1], pcm[0, 0])
    shr_y = np.arctan2(pcm[1, 1], pcm[1, 0]) - np.pi + rot
    rot = apu.Quantity(rot, apu.rad).to(apu.deg)
    shr_y = apu.Quantity(shr_y, apu.rad).to(apu.deg)

    return (sx, sy, rot, shr_y)


def rotate_data(data: np.ndarray, angle: apu.Quantity,
                data_wcs: Optional[apwcs.WCS] = None):
    """
    Rotate data by angle and update the optionally given WCS data.

    Parameters
    ----------
    data : numpy.ndarray
        The data to be rotated.
    angle : apu.Quantity
        The rotation angle.
    data_wcs : astropy.wcs.WCS or None, optional
        A WCS for the data. The default is None.

    Returns
    -------
    dict
        A dictionary with the the following keys and values.

        * 'data' : numpy.ndarray
            The rotated data
        * 'wcs' : astropy.wcs.WCS or None
            The updated WCS for the rotated data
    """
    if data_wcs is not None:
        # ndimage shape is [height, width] and element [0, 0] has
        # pixel coordinates [1, 1]
        rot_center = (np.array((data.shape[1], data.shape[0])) / 2) + 1
        rr = -angle.to(apu.rad).value
        pcm = data_wcs.celestial.wcs.get_pc()
        crpix = np.array(data_wcs.wcs.crpix)

        rot_matrix = np.array(
            [
                [1 * np.cos(rr), -1 * np.sin(rr)],
                [1 * np.sin(rr), 1 * np.cos(rr)]
            ]
        )

        new_pcm = rot_matrix.dot(pcm)
        new_crpix = rot_matrix.dot(crpix - rot_center) + rot_center

        new_wcs = apwcs.WCS(data_wcs.to_header())
        new_wcs.wcs.pc = new_pcm
        new_wcs.wcs.crpix = new_crpix
    else:
        new_wcs = None

    rotated_data = data.copy().astype('float32')
    nan_mask = np.isnan(rotated_data)

    rotated_data = rotate(
        rotated_data,
        angle.to(apu.deg).value,
        reshape=False,
        prefilter=False,
        order=0,
        cval=np.nan
    )

    rotated_mask = rotate(
        nan_mask,
        angle.to(apu.deg).value,
        reshape=False,
        prefilter=False,
        order=0,
        cval=1
    )

    rotated_data[rotated_mask] = np.nan

    return {'data': rotated_data, 'wcs': new_wcs}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Generate a synthetic spectral datacube that can be used to test specex module.

Copyright (C) 2022  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
import os
import sys
import warnings
import numpy as np

from astropy.io import fits
from astropy.modeling import models
from scipy.ndimage import gaussian_filter
from astropy.table import Table
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from astropy.utils.exceptions import AstropyWarning
from astropy import units

from test import TEST_DATA_PATH
from specex.utils import get_sdss_spectral_templates, get_sdss_template_data


class FakeObject():
    """
    Class to handle fake object generation.

    This class contains morphological properties of an object, such as the
    the shape of the isophotal ellipse at kron radius, as well as physical
    properties as its spectral tepmlate and luminosity.
    """

    __template_dict = {}

    def __init__(self, template, x_image, y_image, z=None,
                 surface_bightness_profile=None):
        """
        Initialize the class.

        Parameters
        ----------
        template_file : str
            Path of a SDSS template file.
        x_image : float
            The x position of the center of the object in the image.
        y_image : float
            The y position of the center of the object in the image.
        z : float
            Redshift of the spectrum.
        surface_bightness_profile : astropy.modeling.Fittable2DModel, optional
            The surface brightness profile model. If None, a Sersic2D model is
            used with random ellipticity and theta. The default is None.

        Returns
        -------
        None.

        """
        self.x_image = x_image
        self.y_image = y_image
        self.z = z

        self.template_file = template

        if surface_bightness_profile is None:
            obj_type = self.template_file['type']
            obj_type = obj_type.lower()
            if obj_type == 'star':
                n_val = 99
                amp_val = 1
                e_val = 0
                r_eff_val = 2
                if self.z is None:
                    self.z = 0
            elif obj_type == 'galaxy':
                n_val = 1 + np.random.random()*5
                amp_val = 10 + np.random.random()*5
                e_val = np.random.random()*0.6
                r_eff_val = np.random.normal(loc=2)**2
                if self.z is None:
                    self.z = 0
                    while self.z < 0.2:
                        self.z = np.random.rand() * 2
            else:
                n_val = 99
                amp_val = 5.0e-6
                e_val = 0
                r_eff_val = 1.5
                if self.z is None:
                    self.z = 0
                    while self.z < 1:
                        self.z = np.random.rand() * 5

            self.surface_bightness_profile_function = models.Sersic2D(
                amplitude=amp_val,
                r_eff=r_eff_val,
                n=n_val,
                x_0=self.x_image,
                y_0=self.y_image,
                ellip=e_val,
                theta=np.random.random() * 2 * np.pi
            )
        else:
            self.surface_bightness_profile_function = surface_bightness_profile

    @property
    def template(self):
        """
        Get the spectral template of the object.

        Returns
        -------
        template : dict
            The spectral template dictionary returned by
            specex.utils.get_sdss_template_data().

        """
        return get_sdss_template_data(self.template_file['file'])

    def get_image(self, width, height, seeing=1):
        """
        Get the an image containing the object.

        Parameters
        ----------
        width : int
            The image width.
        height : int
            The image height.
        seeing: float
            Parameter of a gaussian blur filter.

        Returns
        -------
        img : 2D numpy.ndarray
            The image containing the object.

        """
        x, y = np.meshgrid(np.arange(width), np.arange(height))
        base_image = self.surface_bightness_profile_function(x, y)

        base_image += gaussian_filter(base_image, seeing)
        return base_image

    def get_spectrum(self, wave_range, wave_step=1):
        """
        Get the spectrum of the object.

        Parameters
        ----------
        wave_range : tuple or numpy.ndarray
            The range of wavelenght, in Angstrom, the sythetic spectrum should
            cover. Only maximum and minimum values of this parameter are used.

        Returns
        -------
        wave : numpy.ndarray
            Wavelenghts, in Angstrom, used to generate spectral data
        flux : numpy.ndarray
            Fluxes, in arbitrary units, at the different wavelengths.

        """
        wave, flux = gen_fake_spectrum(
            wave_range,
            self.template,
            z=self.z,
            wave_step=wave_step
        )
        return wave, flux

    def get_cube(self, width, height, wave_range,  wave_step=1, seeing=1):
        """
        Get a datacube containing the spectrum of the object.

        Parameters
        ----------
        width : int
            The image width.
        height : int
            The image height.
        wave_range : tuple or numpy.ndarray
            The range of wavelenght, in Angstrom, the sythetic spectrum should
            cover. Only maximum and minimum values of this parameter are used.
        wave_step : float, optional
            Resolution of the spectrum in Angstrom. The default is 1.

        Returns
        -------
        cube : 3D numpy.ndarray
            The spectral datacybe of the object.

        """
        waves, flux = self.get_spectrum(wave_range, wave_step)
        image = self.get_image(width, height, seeing)
        cube = image[..., None] * flux
        noise = np.random.random(size=cube.shape) * 0.005
        cube = cube + noise
        cube = cube.transpose(2, 0, 1)
        return cube


def get_waves(wave_range, wave_step=1.0):
    """
    Generate a wavelength grid given an input wavelenght range.

    Parameters
    ----------
    wave_range : tuple or numpy.ndarray
        The range of wavelenght, in Angstrom, the sythetic spectrum should
        cover. Only maximum and minimum values of this parameter are used.

    Returns
    -------
    wave : numpy.ndarray
        Wavelenghts, in Angstrom.

    """
    # generate a wavelenght grid
    w_start = np.min(wave_range)
    w_end = np.max(wave_range)
    wave = np.arange(w_start, w_end, step=wave_step)
    return wave


def gen_fake_spectrum(wave_range, template, z, wave_step=1.0):
    """
    Generate a synthetic spectrum using a specific template.

    Parameters
    ----------
    wave_range : tuple or numpy.ndarray
        The range of wavelenght, in Angstrom, the sythetic spectrum should
        cover. Only maximum and minimum values of this parameter are used.
    template : SDSS spectral template data
        spectral data dictionary compatible with the format returned by
        specex.get_sdss_template_data()
    z : float
        Redshift of the spectrum.
    wave_step : float, optional
        Resolution of the spectrum in Angstrom. The default is 1.

    Returns
    -------
    wave : numpy.ndarray
        Wavelenghts, in Angstrom, used to generate spectral data
    flux : numpy.ndarray
        Fluxes, in arbitrary units, at the different wavelengths.
    """
    wave = get_waves(wave_range, wave_step)
    flux = np.interp(wave, template['wavelengths'] * (1 + z), template['flux'])
    return wave, flux - np.min(flux)


def gen_fake_cube(out_dir, width, height, wave_range, n_objects, wave_step=1.0,
                  seeing=1, overwrite=False):

    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    templates = get_sdss_spectral_templates(out_dir, use_cached=True)

    reg_file = os.path.join(out_dir, 'test_cube.reg')
    cat_file = os.path.join(out_dir, 'test_cube_cat.fits')
    cube_file = os.path.join(out_dir, 'test_cube.fits')

    out_files_list = [reg_file, cat_file, cube_file]

    if all([os.path.isfile(x) for x in out_files_list]):
        if not overwrite:
            return out_files_list

    print(f"Generating a synthetic cube to {cube_file}")

    waves = get_waves(wave_range, wave_step)

    header = fits.Header()

    header['BUNIT'] = '10**(-20)*erg/s/cm**2/Angstrom'
    header['CRPIX1'] = 135.5
    header['CRPIX2'] = 132.0
    header['CD1_1'] = 2.86146E-06
    header['CD1_2'] = -2.07292E-05
    header['CD2_1'] = -2.09168E-05
    header['CD2_2'] = -2.83578E-06
    header['CUNIT1'] = 'deg     '
    header['CUNIT2'] = 'deg     '
    header['CTYPE1'] = 'RA---TAN'
    header['CTYPE2'] = 'DEC--TAN'
    header['CRVAL1'] = 182.6361425552
    header['CRVAL2'] = 39.40589479746

    my_wcs = WCS(header)

    w_padding = width/50
    h_padding = width/50

    obj_attributes = zip(
        w_padding + np.random.random(n_objects) * (width - 2*w_padding),
        h_padding + np.random.random(n_objects) * (height - 2*h_padding),
        np.random.choice(templates, n_objects)
    )

    objects = [
        FakeObject(
            attr[2],
            attr[0], attr[1],
        )
        for attr in obj_attributes
    ]

    base_cube = np.zeros((waves.shape[0], height, width))
    for j, obj in enumerate(objects):
        try:
            base_cube += obj.get_cube(
                width, height,
                wave_range, wave_step,
                seeing
            )
        except ValueError:
            print(f"WARNING: Skipping object {j}")
            continue
        else:
            sys.stdout.write(f"\rprogress: {j/ len(objects):.2%}\r")
            sys.stdout.flush()

    print("DONE!")

    myt = Table(
        names=[
            'NUMBER',
            'X_IMAGE',
            'Y_IMAGE',
            'A_IMAGE',
            'B_IMAGE',
            'THETA_IMAGE',
            'KRON_RADIUS',
            'ALPHA_J2000',
            'DELTA_J2000',
            'TRUE_Z',
            'TRUE_TYPE',
            'TRUE_SUBTYPE'
        ],
        dtype=[
            int, float, float, float, float, float, float, float, float, float,
            'U10', 'U10'
        ]
    )

    celestial_wcs = my_wcs.celestial
    pixelscale_x, pixelscale_y = [
        units.Quantity(sc_val, u)
        for sc_val, u in zip(
            proj_plane_pixel_scales(celestial_wcs),
            celestial_wcs.wcs.cunit
        )
    ]

    regfile_text = 'fk5\n'
    for i, obj in enumerate(objects):
        obj_x = int(obj.x_image)
        obj_y = int(obj.y_image)
        obj_kron = obj.surface_bightness_profile_function.r_eff.value
        e_val = obj.surface_bightness_profile_function.ellip.value
        obj_a_image = 3 * obj_kron * pixelscale_x
        obj_a_image = obj_a_image.to('arcsec').value
        obj_b_image = 3 * obj_kron * (1 - e_val) * pixelscale_y
        obj_b_image = obj_b_image.to('arcsec').value
        obj_theta_image = obj.surface_bightness_profile_function.theta.value
        obj_theta_image = np.rad2deg(obj_theta_image)

        sky_coords = my_wcs.pixel_to_world(obj.x_image, obj.y_image)

        obj_ra = sky_coords.ra.deg
        obj_dec = sky_coords.dec.deg

        regfile_text += f"ellipse({obj_ra:.6f}, {obj_dec:.6f}, "
        regfile_text += f"{obj_a_image:.4f}\", "
        regfile_text += f"{obj_b_image:.4f}\", {obj_theta_image:.4f}) "
        regfile_text += f"# text={{S{i:06d}}}\n"

        new_row = [
            i, obj_x, obj_y, obj_a_image, obj_b_image, obj_theta_image,
            obj_kron, obj_ra, obj_dec, obj.z, obj.template_file['type'],
            obj.template_file['sub-type']
        ]

        myt.add_row(new_row)

    with open(reg_file, "w") as f:
        f.write(regfile_text)

    myt.write(cat_file, overwrite=True)

    header['CRVAL3'] = waves[0]
    header['CRPIX3'] = 1.0
    header['CTYPE3'] = 'WAVE'
    header['CD3_3'] = wave_step
    header['CD1_3'] = 0.0
    header['CD2_3'] = 0.0
    header['CD3_1'] = 0.0
    header['CD3_2'] = 0.0
    header['CUNIT3'] = 'Angstrom'
    header['OBJECT'] = 'SYNTHETIC'
    header['PCOUNT'] = 0
    header['GCOUNT'] = 1

    cube_hdu = fits.ImageHDU(
        data=base_cube,
        name='SPECTRUM',
        header=header
    )

    cube_hdu.writeto(cube_file, overwrite=True)
    return out_files_list


def main(overwrite=False):
    """
    Run the main program of this module.

    Returns
    -------
    None.

    """
    warnings.simplefilter('ignore', category=AstropyWarning)

    wave_range = (4500, 8000)

    return gen_fake_cube(
        TEST_DATA_PATH, 256, 256, wave_range, n_objects=30, wave_step=1
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPEX - SPectra EXtractor.

Extract spectra from spectral data cubes.

Copyright (C) 2022  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
from __future__ import absolute_import, division, print_function

import os
import sys
import unittest
import warnings
from urllib import request

from astropy.io import fits
from astropy.table import Table, join

from specex.cube import cutout_main, smoothing_main
from specex.stack import cube_stack
from specex.sources import detect_from_cube
from specex.zeropoints import main as zpinfo
from specex.cubex import specex
from specex.plot import plot_spectra
from specex.utils import get_pbar

from test import make_synt_cube, get_hst_test_images, TEST_DATA_PATH
from test import make_synt_specs, get_muse_test_cube


try:
    from specex.rrspecex import rrspecex
except ImportError:
    HAS_RR = False
else:
    HAS_RR = True


Z_FTOL = 0.01


NICMOS_REGFILE_DATA = """
# Region file format: DS9 version 4.1
fk5
box(182.6350214,39.4064356,5.000",3.000",262.21022)
circle(182.6356792,39.4058053,2.500")
ellipse(182.6349624,39.4062180,2.500",1.500",217.21022)
"""


class MyTests(unittest.TestCase):

    test_hst_imgs = get_hst_test_images()
    cube_file = get_muse_test_cube()
    spec_files = make_synt_specs.main()
    reg_file = os.path.join(TEST_DATA_PATH, 'test_regions.reg')
    cat_file = os.path.join(TEST_DATA_PATH, 'test_sources.cat')

    def test_zeropoint_info(self):
        print(">>> Testing zeropoint_info...\n")
        if not self.test_hst_imgs:
            print(
                "Failed to download HST test images, skipping this test...",
                file=sys.stderr
            )
            return True
        zpinfo(self.test_hst_imgs)

    def test_cube_stack(self):
        print(">>> Testing cube_stack...\n")
        print(self.cube_file)
        cube_stack_options = [
            self.cube_file,
            '--var-hdu', 'STAT'
        ]
        cube_stack(cube_stack_options)

    def test_cube_smoothing_phisical_units(self):
        print(">>> Testing smoothing...\n")
        smoohting_options = [
            '--verbose',
            '--spatial-sigma', '0.5arcsec,0.25arcsec',
            '--wave-sigma', '1angstrom',
            self.cube_file
        ]
        smoothing_main(smoohting_options)

    def test_cube_smoothing_pixel_units(self):
        print(">>> Testing smoothing...\n")
        smoohting_options = [
            '--verbose',
            '--spatial-sigma', '3,2',
            '--wave-sigma', '2',
            self.cube_file
        ]
        smoothing_main(smoohting_options)

    def test_grayscale_cutout(self):
        print(">>> Testing grayscale cutout...\n")
        if not self.test_hst_imgs:
            print(
                "Failed to download HST test images, skipping this test...",
                file=sys.stderr
            )
            return True
        for img in self.test_hst_imgs:
            if img.endswith('NICMOSn4hk12010_mos.fits'):
                regfile = os.path.join(
                    TEST_DATA_PATH, 'NICMOSn4hk12010_mos.reg'
                )

                with open(regfile, 'w') as f:
                    f.write(NICMOS_REGFILE_DATA)

                cutout_options = [
                    '--verbose',
                    '--regionfile', regfile,
                    img
                ]
                cutout_main(cutout_options)
                break

    def test_cube_cutout(self):
        print(">>> Testing cube cutout...\n")

        cutout_options = [
            '--verbose',
            '--center', '16.3867deg,-24.6464deg',
            '--size', '6arcsec,3arcsec',
            '--angle', '45deg',
            self.cube_file
        ]
        cutout_main(cutout_options)

    def test_cube_cutout_regfile(self):
        print(">>> Testing cube cutout...\n")

        cutout_options = [
            '--verbose',
            '--regionfile', self.reg_file,
            self.cube_file
        ]
        cutout_main(cutout_options)

    def test_extract_sources(self):
        print(">>> Testing sources detection from datacube...\n")
        detect_from_cube([self.cube_file])

    def test_specex_catalog(self):
        print(">>> Testing specex catalog...\n")
        specex_options = [
            '--catalog', self.cat_file,
            '--mode', 'circular_aperture',
            '--aperture-size', '0.8arcsec',
            '--weighting', 'whitelight',
            '--no-nans', self.cube_file
        ]
        specex(options=specex_options)

    def test_specex_regionfile(self):
        print(">>> Testing specex regionfile...")
        specex_options = [
            '--regionfile', self.reg_file,
            '--mode', 'kron_ellipse',
            '--no-nans', self.cube_file
        ]
        specex(options=specex_options)


    def test_specex_regionfile_wlight(self):
        print(">>> Testing specex regionfile (cube whitelight)...")
        specex_options = [
            '--regionfile', self.reg_file,
            '--mode', 'kron_ellipse',
            '--weighting', 'whitelight',
            '--no-nans', self.cube_file
        ]
        specex(options=specex_options)

    def test_specex_regionfile_log_wlight(self):
        print(">>> Testing specex regionfile (cube log-whitelight)...")
        specex_options = [
            '--regionfile', self.reg_file,
            '--mode', 'kron_ellipse',
            '--weighting', 'log-whitelight',
            '--no-nans', self.cube_file
        ]
        specex(options=specex_options)

    def test_specex_regionfile_wimg(self):
        print(">>> Testing specex regionfile (ext. image)...")
        specex_options = [
            '--regionfile', self.reg_file,
            '--mode', 'kron_ellipse',
            '--weighting', "ADP.2023-09-01T12:56:41_data.fits",
            '--no-nans', self.cube_file
        ]
        specex(options=specex_options)

    def test_plot(self):
        print(">>> Testing specex-plot...")

        plot_options = [
            '--restframe',
            '--outdir', 'test_plot_out',
            *self.spec_files
        ]

        plot_spectra(options=plot_options)


"""
# unittest seems to have some issues with redrock implementation...
class TestRRSpex(unittest.TestCase):

    @unittest.skipIf(not HAS_RR, "redrock not installed")
    def test_rrspecex_success(self):
        global HAS_RR

        if not HAS_RR:
            return

        test_files = make_synt_specs.main()

        true_z_table = Table(
            names=['SPECID', 'TRUE_Z'],
            dtype=['U10', 'float32']
        )

        for file in test_files:
            header = fits.getheader(file, ext=0)
            true_z_table.add_row([header['ID'], header['OBJ_Z']])

        options = ['--quite', ] + test_files
        targets, zbest, scandata = rrspecex(options=options)

        zbest = join(true_z_table, zbest, keys=['SPECID'])
        print(zbest)

        for i, obj in enumerate(zbest):
            delta_z = abs(obj['TRUE_Z'] - obj['Z'])/(1 + obj['TRUE_Z'])
            if delta_z >= Z_FTOL:
                warnings.warn(
                    f"OBJ {i}: computed redshift outside f01 limit!",
                )
"""

if __name__ == '__main__':
    tests = MyTests()
    tests.test_zeropoint_info()

    tests.test_stack_cube()

    tests.test_grayscale_cutout()
    tests.test_cube_cutout()

    tests.test_extract_sources()

    tests.test_specex_catalog_success()
    tests.test_specex_regionfile_success()

    tests.test_plot_success()

    # if HAS_RR:
    #     test_06 = TestRRSpex()
    #     test_06.test_rrspecex_success()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

redrock wrapper tools for python-specex spectra.
This program is based on the structure of redrock.external.boss function.

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import traceback
import argparse
from glob import glob
from packaging import version

import numpy as np
from scipy import sparse

from astropy.io import fits
from astropy.table import Table, join
import astropy.wcs as wcs
import matplotlib.pyplot as plt

from scipy.signal import savgol_filter

try:
    import redrock
    from redrock.utils import elapsed, get_mp
    from redrock.targets import Spectrum, Target, DistTargetsCopy
    from redrock.templates import load_dist_templates, find_templates, Template
    from redrock.results import write_zscan
    from redrock.zfind import zfind
    from redrock._version import __version__
    from redrock.archetypes import All_archetypes
except (ModuleNotFoundError, Exception):
    HAS_REDROCK = False
else:
    HAS_REDROCK = True

from specex.exceptions import exception_handler
from specex.utils import (
    plot_zfit_check, plot_scandata, get_pbar
)

from specex.cube import (
    KNOWN_SPEC_EXT_NAMES,
    KNOWN_VARIANCE_EXT_NAMES,
    KNOWN_INVAR_EXT_NAMES,
    KNOWN_MASK_EXT_NAMES,
    KNOWN_RCURVE_EXT_NAMES
)

MUSE_RESOLUTION_ANG = 2.51
RR_GPU_MIN_VER = "0.16.0"

def get_templates(template_types=[], filepath=False, templates=None):
    """
    Get avilable templates.

    Parameters
    ----------
    template_types : list of str, optional
        List of template types to retrieve. If it's empty all available
        templates will be returned.
        The default is [].
    filepath : boot, optional
        If it's true then return the file paths instead of actual templates.
    templates : str, optional
        The path of a template file or of a directory containing templates
        files. If None, templates are searched in the default redrock path.
        The default value is None.

    Returns
    -------
    available_templates : list of redrock.templates.Template or file paths
        The available templates or the corresponding file paths.

    """
    if templates is not None and os.path.isfile(templates):
        return [Template(templates), ]

    available_templates = []
    for t in find_templates(templates):
        templ = Template(t)
        if not template_types or templ.template_type in template_types:
            if filepath:
                available_templates.append(t)
            else:
                available_templates.append(templ)

    return available_templates


def get_template_types():
    """
    Get the available types of templates.

    Returns
    -------
    types : list of str
        List of types of available templates.

    """
    templates = [
        t.template_type
        for t in get_templates()
    ]
    types = set(templates)
    return types


def write_zbest(outfile, zbest, template_version, archetype_version):
    """
    Write zbest Table to outfile.

    Parameters
    ----------
        outfile : str
            The output file path.
        zbest : astropy.table.Table
            The output redshift fitting results.

    """
    header = fits.Header()
    header['RRVER'] = (__version__, 'Redrock version')

    for i, fulltype in enumerate(template_version.keys()):
        header['TEMNAM'+str(i).zfill(2)] = fulltype
        header['TEMVER'+str(i).zfill(2)] = template_version[fulltype]

    if archetype_version is not None:
        for i, fulltype in enumerate(archetype_version.keys()):
            header['ARCNAM'+str(i).zfill(2)] = fulltype
            header['ARCVER'+str(i).zfill(2)] = archetype_version[fulltype]

    zbest.meta['EXTNAME'] = 'ZBEST'

    hx = fits.HDUList()
    hx.append(fits.PrimaryHDU(header=header))
    hx.append(fits.convenience.table_to_hdu(zbest))
    hx.writeto(os.path.expandvars(outfile), overwrite=True)
    return


def read_spectra(spectra_fits_list, spec_hdu=None, var_hdu=None, wd_hdu=None,
                 memmap=True, resolution=2.51, smoothing=0, quite=False):
    """
    Read input spectra fits files.

    Parameters
    ----------
    spectra_fits_list : list
        List of fits files containing the input spectra.
    spec_hdu : int or None, optional
        The index of the HDU that contains the spectral data itself.
        If it is None, then the index is determined automatically by the name
        of the HDU. If this operation fails ValueError exception is raised.
        The default value is None.
    var_hdu : int or None, optional
        The index of the HDU that contains the  variance of the spectral data.
        If it is None, then the index is determined automatically by the name
        of the HDU. If this operation fails ValueError exception is raised.
        The default value is None.
    wd_hdu : int or None, optional
        The index of the HDU that contains the wavelength dispersion in pixels.
        If it is None, then the index is determined automatically by the name
        of the HDU. If this operation fails, no wavelenght dispersion will be
        used and the spectra will be considered having a uniform resolution.
        The default value is None.
    memmap : bool, optional
        Whether to use memmap or not. The default value is False.
    resolution : float, optional
        A fixed spectral resolution in Angrstrom to be used if no dispersion
        information is available from the spectrum file itself. The default
        value is 2.51 (the MUSE resolution).
    smoothing : int, optional
        A smooting to be applyed before redshift estimation.
        A value of 0 means no smoothing. The default value is 0.
    quite : bool, optional
        Reduce the verbosity of the output.

    Raises
    ------
    ValueError
        If cannot automatically determine the HDU containing the specral data
        or its variance.

    Returns
    -------
    targets : list of redrock.targets.Target
        The target spectra for which redshift will be computed.
    metatable : astropy.table.Table
        A table containing metadata.

    """
    targets = []
    targetids = []
    target_file_names = []
    specids = []
    sn_vals = []
    sn_var_vals = []
    sn_em_vals = []

    # crop to templates limits
    lmin = 3500.
    lmax = 10000.

    for j, fits_file in enumerate(spectra_fits_list):
        if not quite:
            progress = (j + 1) / len(spectra_fits_list)
            sys.stderr.write(f"\r{get_pbar(progress)} {progress:.2%}\r")
            sys.stderr.flush()
        hdul = fits.open(fits_file, memmap=memmap)

        valid_id_keys = [
            f"{i}{j}"
            for i in ['', 'OBJ', 'OBJ_', 'TARGET', 'TARGET_']
            for j in ['ID', 'NUMBER', 'UID', 'UUID']
        ]

        target_id = f"{j:09}"
        target_file_names.append(fits_file)
        spec_id = target_id
        for hdu in hdul:
            for key in valid_id_keys:
                try:
                    spec_id = hdu.header[key]
                except KeyError:
                    continue
                else:
                    break

        if spec_hdu is None:
            for hdu in hdul:
                if hdu.name.lower() in KNOWN_SPEC_EXT_NAMES:
                    flux = hdu.data
                    if smoothing > 0:
                        window_size = 4*smoothing + 1
                        flux = savgol_filter(flux, window_size, 3)
                    spec_header = hdu.header
                    spec_wcs = wcs.WCS(spec_header)
                    break
            else:
                raise ValueError(
                    "Cannot determine the HDU containing spectral data: "
                    f"'{fits_file}'"
                )
        else:
            flux = hdul[spec_hdu].data
            spec_wcs = wcs.WCS(hdul[spec_hdu].header)

        for hdu in hdul:
            if hdu.name.lower() in KNOWN_MASK_EXT_NAMES:
                nanmask = hdu.data.astype(bool)
                break
        else:
            nanmask = None

        if var_hdu is None:
            for hdu in hdul:
                if hdu.name.lower() in KNOWN_VARIANCE_EXT_NAMES:
                    ivar = 1 / hdu.data
                    break
                elif hdu.name.lower() in KNOWN_INVAR_EXT_NAMES:
                    ivar = hdu.data
                    break
            else:
                print(
                    "WARNING: Cannot determine the HDU containing variance "
                    f"data in '{fits_file}'! Using dumb constan variance...",
                )
                ivar = np.ones_like(flux)
        else:
            ivar = 1 / hdul[var_hdu].data

        if wd_hdu is None:
            for hdu in hdul:
                if hdu.name.lower() in KNOWN_RCURVE_EXT_NAMES:
                    wd = hdu.data
                    break
            else:
                wd = None
        else:
            wd = hdul[wd_hdu].data

        if flux.shape != ivar.shape:
            raise ValueError(
                f"'{fits_file}' - "
                "Spectral data invalid or corruptede: Flux data shape "
                "do not match variance data one!"
            )

        main_header = hdul[0].header

        # NOTE: Wavelenghts must be in Angstrom units
        pixel = np.arange(len(flux))
        if spec_wcs.has_spectral:
            lam = spec_wcs.pixel_to_world(pixel).Angstrom
        else:
            try:
                coeff0 = spec_header["COEFF0"]
                coeff1 = spec_header["COEFF1"]
            except KeyError:
                continue
            lam = 10**(coeff0 + coeff1*pixel)
        flux = flux.astype('float32')

        if nanmask is None:
            flux_not_nan_mask = ~np.isnan(flux)
        else:
            flux_not_nan_mask = ~nanmask

        flux = flux[flux_not_nan_mask]
        ivar = ivar[flux_not_nan_mask]
        lam = lam[flux_not_nan_mask]
        if wd is not None:
            wd = wd[flux_not_nan_mask]
        else:
            # If now wavelenght dispersion information is present, then
            # compute it using the wavelenght
            delta_lambda = np.ones_like(lam)
            delta_lambda[1:] = (lam[1:] - lam[:-1])
            wd = resolution / delta_lambda
            wd[0] = wd[1]
        wd[wd < 1e-3] = 2.

        imin = abs(lam-lmin).argmin()
        imax = abs(lam-lmax).argmin()

        lam = lam[imin:imax]
        flux = flux[imin:imax]
        ivar = ivar[imin:imax]
        wd = wd[imin:imax]

        ndiag = int(4*np.ceil(wd.max())+1)
        nbins = wd.shape[0]

        ii = np.arange(lam.shape[0])
        di = ii-ii[:, None]
        di2 = di**2

        # build resolution from wdisp
        reso = np.zeros([ndiag, nbins])

        for idiag in range(ndiag):
            offset = ndiag//2-idiag
            d = np.diagonal(di2, offset=offset)
            if offset < 0:
                reso[idiag, :len(d)] = np.exp(-d/2/wd[:len(d)]**2)
            else:
                reso[idiag, nbins-len(d):nbins] = np.exp(
                    -d/2/wd[nbins-len(d):nbins]**2
                )

        reso /= np.sum(reso, axis=0)
        offsets = ndiag//2 - np.arange(ndiag)
        nwave = reso.shape[1]
        R = sparse.dia_matrix((reso, offsets), (nwave, nwave))

        try:
            s_n = main_header['SN']
        except KeyError:
            s_n = -1

        try:
            s_n_var = main_header['SN_VAR']
        except KeyError:
            s_n_var = -1

        try:
            s_n_em = main_header['SN_EMISS']
        except KeyError:
            s_n_em = -1

        rrspec = Spectrum(lam, flux, ivar, R, None)
        target = Target(target_id, [rrspec])
        target.input_file = fits_file
        target.spec_id = spec_id
        targets.append(target)
        targetids.append(target_id)
        specids.append(spec_id)
        sn_vals.append(s_n)
        sn_var_vals.append(s_n_var)
        sn_em_vals.append(s_n_em)

    metatable = Table()
    metatable['TARGETID'] = targetids
    metatable['SPECID'] = specids
    metatable['FILE'] = target_file_names
    metatable['SN'] = sn_vals
    metatable['SN_VAR'] = sn_var_vals
    metatable['SN_EMISS'] = sn_em_vals

    if not quite:
        print("", file=sys.stderr)

    return targets, metatable


def __argshandler(options=None):
    """
    Handle input arguments.

    Parameters
    ----------
    options : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    args : TYPE
        DESCRIPTION.

    """
    parser = argparse.ArgumentParser(
        description="Estimate redshifts for spectra extracted with"
        "python-specex using redrock interface."
    )

    parser.add_argument(
        "spectra", metavar='SPECTRA_FITS', type=str, nargs='+',
        help="input spectra fits files"
    )

    parser.add_argument(
        "-t", "--templates", type=str, default=None, required=False,
        help="template file or directory"
    )

    parser.add_argument(
        "--archetypes", type=str, default=None, required=False,
        help="archetype file or directory for final redshift comparisons"
    )

    parser.add_argument(
        "-o", "--output", type=str, default=None, required=False,
        help="output file"
    )

    parser.add_argument(
        "--zbest", type=str, default=None, required=False,
        help="output zbest FITS file"
    )

    parser.add_argument(
        "--gpu", action="store_true", default=False,
        help="Use GPU computing acceleration (requires redrock >= "
        f"{RR_GPU_MIN_VER})."
    )

    parser.add_argument(
        "--priors", type=str, default=None, required=False,
        help="optional redshift prior file"
    )

    parser.add_argument(
        "--chi2-scan", type=str, default=None, required=False,
        help="Load the chi2-scan from the input file"
    )

    parser.add_argument(
        "--nminima", type=int, default=3, required=False,
        help="the number of redshift minima to search"
    )

    parser.add_argument(
        "--mp", type=int, default=1, required=False,
        help="if not using MPI, the number of multiprocessing processes to use"
        " (defaults to half of the hardware threads)"
    )

    parser.add_argument(
        "--debug", default=False, action="store_true", required=False,
        help="debug with ipython (only if communicator has a single process)"
    )

    parser.add_argument(
        "--plot-zfit", action="store_true", default=False, required=False,
        help="Generate plots of the spectra with infomrazion about the "
        "result of the template fitting (i.e. the best redshift, the position "
        "of most important lines, the best matching template, etc...)."
    )

    parser.add_argument(
        "--checkimg-outdir", type=str, default='checkimages', required=False,
        help='Set the directory where check images are saved (when they are '
        'enabled thorugh the appropriate parameter).'
    )

    parser.add_argument(
        "--quite", action='store_true', default=False,
        help="Reduce program output at bare minimum. Do not print fitting "
        "result,"
    )

    parser.add_argument(
        "--no-memmap", action='store_true', default=False,
        help='Disable memmapping to speed the computation at a cost of an '
        'increased memory usage.'
    )

    parser.add_argument(
        "--resolution", type=float, default=2.51, help="A fixed resolution of "
        "the spectrograph to be used if no resolution curve is present in the "
        "input spectrum itself."
    )

    parser.add_argument(
        '--smoothing', metavar='WINDOW_SIZE', type=int,  default=0,
        help='If %(metavar)s >= 0, then the spectrua are smoothed before the '
        'redshift estimation procedure is run.  %(metavar)s = 0 means that '
        'no smoothing is applyed. If not specified, the default '
        '%(metavar)s = %(default)s is used.'
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    return args


def rrspecex(options=None, comm=None):
    """
    Estimate redshifts for spectra extracted with python-specex using redrock.

    This loads targets serially and copies them into a DistTargets class.
    It then runs redshift fitting and writes the output to a catalog.

    Parameters
    ----------
    options : list, optional
        lLst of commandline options to parse. The default is None.
    comm : mpi4py.Comm, optional
        MPI communicator to use. The default is None.

    Returns
    -------
    targets : list of redrock.targets.Target objects
        list of target spectra.
    zfit : astropy.table.Table
        Table containing the fit results.
    scandata : dict
        A dictionary containing the redshift scanning information for each
        target
    """
    global HAS_REDROCK
    global HAS_IPYTHON

    global_start = elapsed(None, "", comm=comm)
    comm_size = 1
    comm_rank = 0
    if comm is not None:
        comm_size = comm.size
        comm_rank = comm.rank

    args = __argshandler(options)

    if args.debug:
        sys.excepthook = exception_handler

    # Check arguments - all processes have this, so just check on the first
    # process
    if comm_rank == 0:
        if args.debug and comm_size != 1:
            print(
                "--debug can only be used if the communicator has one process"
            )
            sys.stdout.flush()
            if comm is not None:
                comm.Abort()

        if (args.output is None) and (args.zbest is None):
            print("--output or --zbest required")
            sys.stdout.flush()
            if comm is not None:
                comm.Abort()

    # Multiprocessing processes to use if MPI is disabled.
    mpprocs = 0
    if comm is None:
        mpprocs = get_mp(args.mp)
        print("Running with {} processes".format(mpprocs))
        if "OMP_NUM_THREADS" in os.environ:
            nthread = int(os.environ["OMP_NUM_THREADS"])
            if nthread != 1:
                print("WARNING:  {} multiprocesses running, each with "
                      "{} threads ({} total)".format(
                          mpprocs, nthread, mpprocs*nthread
                      ))
                print("WARNING:  Please ensure this is <= the number of "
                      "physical cores on the system")
        else:
            print("WARNING:  using multiprocessing, but the OMP_NUM_THREADS")
            print("WARNING:  environment variable is not set- your system may")
            print("WARNING:  be oversubscribed.")
        sys.stdout.flush()
    elif comm_rank == 0:
        print("Running with {} processes".format(comm_size))
        sys.stdout.flush()

    try:
        # Load and distribute the targets
        if comm_rank == 0:
            print("Loading targets...")
            sys.stdout.flush()

        start = elapsed(None, "", comm=comm)

        # Read the spectra on the root process.  Currently the "meta" Table
        # returned here is not propagated to the output zbest file.  However,
        # that could be changed to work like the DESI write_zbest() function.
        # Each target contains metadata which is propagated to the output zbest
        # table though.

        # Windows prompt does not expand globs, so let's do it
        spectra_list = []
        for globbed_fname in args.spectra:
            for fname in glob(globbed_fname):
                spectra_list.append(fname)

        targets, meta = read_spectra(
            spectra_list,
            memmap=not args.no_memmap,
            resolution=args.resolution,
            smoothing=args.smoothing,
            quite=args.quite
        )

        if len(targets) == 0:
            raise ValueError("No spectra were loaded!")
        else:
            print(f"Loaded {len(targets)} spectra")

        _ = elapsed(
            start, "Read of {} targets".format(len(targets)), comm=comm
        )

        # Distribute the targets.

        start = elapsed(None, "", comm=comm)

        dtargets = DistTargetsCopy(targets, comm=comm, root=0)

        # Get the dictionary of wavelength grids
        dwave = dtargets.wavegrids()

        _ = elapsed(
            start,
            "Distribution of {} targets".format(len(dtargets.all_target_ids)),
            comm=comm
        )

        opt_zfind_args = {}
        opt_load_dist_templates_args = {}
        if version.parse(redrock.__version__) >= version.parse(RR_GPU_MIN_VER):
            opt_zfind_args['use_gpu'] = args.gpu
            opt_load_dist_templates_args['use_gpu'] = args.gpu

        # Read the template data
        dtemplates = load_dist_templates(
            dwave,
            templates=args.templates,
            comm=comm,
            mp_procs=mpprocs,
            **opt_load_dist_templates_args
        )

        # Compute the redshifts, including both the coarse scan and the
        # refinement.  This function only returns data on the rank 0 process.
        start = elapsed(None, "", comm=comm)
        scandata, zfit = zfind(
            dtargets,
            dtemplates,
            mpprocs,
            nminima=args.nminima,
            archetypes=args.archetypes,
            priors=args.priors,
            chi2_scan=args.chi2_scan,
            **opt_zfind_args
        )

        _ = elapsed(start, "Computing redshifts took", comm=comm)

        # Write the outputs
        if args.output is not None:
            start = elapsed(None, "", comm=comm)
            if comm_rank == 0:
                write_zscan(args.output, scandata, zfit, clobber=True)
            _ = elapsed(start, "Writing zscan data took", comm=comm)

        # Change to upper case like DESI
        for colname in zfit.colnames:
            if colname.islower():
                zfit.rename_column(colname, colname.upper())

        matched_zfit = join(zfit, meta, keys=['TARGETID'], join_type='left')

        zbest = None
        if comm_rank == 0:
            zbest = matched_zfit[matched_zfit['ZNUM'] == 0]
            zbest.add_index('TARGETID')

            if args.plot_zfit:

                if not os.path.isdir(args.checkimg_outdir):
                    os.makedirs(args.checkimg_outdir)

                available_templates = get_templates(
                    templates=args.templates
                )
                for target in targets:
                    orig_file_name = os.path.basename(
                        zbest.loc[target.id]['FILE']
                    )

                    fig, ax = plot_zfit_check(
                        target,
                        zbest,
                        plot_template=available_templates
                    )
                    figname = f'rrspecex_{orig_file_name}.png'
                    if args.checkimg_outdir is not None:
                        figname = os.path.join(args.checkimg_outdir, figname)
                    fig.savefig(figname, dpi=150)
                    plt.close(fig)

                    if args.debug:
                        figname = f'rrspecex_scandata_{orig_file_name}.png'
                        figname = os.path.join(args.checkimg_outdir, figname)
                        fig, axs = plot_scandata(target, scandata)
                        fig.savefig(figname, dpi=150)
                        plt.close(fig)

            zbest.remove_column('TARGETID')

            if args.zbest:
                start = elapsed(None, "", comm=comm)
                if comm_rank == 0:

                    # Remove extra columns not needed for zbest
                    # zbest.remove_columns(['zz', 'zzchi2', 'znum'])
                    # zbest.remove_columns(['ZNUM'])

                    template_version = {
                        t._template.full_type: t._template._version
                        for t in dtemplates
                    }

                    archetype_version = None
                    if args.archetypes is not None:
                        archetypes = All_archetypes(
                            archetypes_dir=args.archetypes
                        ).archetypes
                        archetype_version = {
                            name: arch._version
                            for name, arch in archetypes.items()
                        }

                    write_zbest(
                        args.zbest, zbest, template_version, archetype_version
                    )

                _ = elapsed(start, "Writing zbest data took", comm=comm)

            if (not args.zbest) and (args.output is None) and (not args.quite):
                print("")
                print(zbest)
                print("")

    except Exception:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = [
            "Proc {}: {}".format(comm_rank, x)
            for x in lines
        ]
        print("".join(lines))
        sys.stdout.flush()
        if comm is not None:
            comm.Abort()

    _ = elapsed(global_start, "Total run time", comm=comm)

    return targets, zbest, scandata


if __name__ == '__main__':
    _ = rrspecex()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

This module provides functions to stack spectral cubes.

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
import os
import argparse
from typing import Union

import numpy as np
from astropy.wcs import WCS
from astropy.io import fits
import matplotlib.pyplot as plt

from specex.utils import stack
from specex.cube import get_hdu


def stack_and_plot(ext, basename, suffix="", is_mask=False, override_wcs=None,
                   dpi=150, wave_ranges=None):
    """
    Stack and plot a spectral datacube.

    Stack a datacube along the spectral axis and plot the result as a
    PNG and a FITS file.

    Parameters
    ----------
    ext : astropy.io.fits.ImageHDU
        The fits exension containing the datacube.
    basename : str
        The name of the output images.
    suffix : str, optional
        An optional suffix for the output image names. The default is "".
    is_mask : bool, optional
        If True, the stacked image is trated as a boolean mask where every
        pixel having non-zero value will be considered as True.
        The default is False.
    override_wcs : astropy.wcs.WCS, optional
        An optional WCS to override the one present in the datacube.
        The default is None.
    dpi : int, optional
        The resolution of the output PNG file. The default is 150.
    wave_range : TYPE, optional
        The optional wavelength range to use, if None all the available
        wavelenghts are used. The default is None.

    Returns
    -------
    new_data : numpy.ndarray
        The stacked image.
    img_wcs : astropy.wcs.WCS
        The WCS of the stacked image.

    """
    if ext.data is None:
        return None

    img_wcs = WCS(ext.header)

    wave_mask: Union[np.ndarray, None]
    if wave_ranges is None:
        wave_mask = None
    elif img_wcs.has_spectral:
        wave_mask = np.zeros(ext.data.shape[0], dtype=bool)
        for wave_range in wave_ranges:
            wave_index = np.arange(ext.data.shape[0])
            wave_angstrom = img_wcs.spectral.pixel_to_world(wave_index)
            wave_angstrom = wave_angstrom.Angstrom
            mask = wave_angstrom >= np.nanmin(wave_range)
            mask &= wave_angstrom <= np.nanmax(wave_range)
            wave_mask |= mask


    img_height, img_width = ext.data.shape[1], ext.data.shape[2]
    img_figsize = (
        img_width / dpi,
        img_height / dpi
    )

    new_data = stack(ext.data, wave_mask=wave_mask)

    if is_mask:
        new_data = ~(new_data == new_data.max()) * 1.0

    levels = np.percentile(new_data[np.isfinite(new_data)], [3, 97])
    fig, ax = plt.subplots(
        1, 1,
        figsize=img_figsize,
        subplot_kw={'projection': img_wcs.celestial}
    )

    ax.imshow(new_data, vmin=np.min(levels), vmax=np.max(levels))
    out_name_png = f"{basename}_{suffix}.png"
    fig.savefig(out_name_png, dpi=dpi, bbox_inches="tight")
    plt.close(fig)

    if override_wcs:
        new_header = override_wcs.celestial.to_header()
    else:
        new_header = img_wcs.celestial.to_header()
    out_name_fits = f"{basename}_{suffix}.fits"
    hdu = fits.PrimaryHDU(data=new_data, header=new_header)
    hdu.writeto(out_name_fits, overwrite=True)
    return new_data, img_wcs


def __argshandler(options=None):
    """
    Parse the arguments given by the user.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument(
        'input_cube', metavar='SPEC_CUBE', type=str, nargs=1,
        help='The spectral cube in fits format from which spectra will be '
        'extracted.'
    )
    parser.add_argument(
        '--dpi', metavar='DPI', type=int, default=150,
        help='Set the DPI of the output figures. '
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--spec-hdu', metavar='SPEC_HDU', type=str, default=-1,
        help='Index or name of the HDU containing the spectral data to '
        'use. Set this to -1 to automatically detect the HDU containing '
        'the fluxes. '
        'The default value is %(metavar)s=%(default)s. '
        'NOTE that this value is zero indexed (i.e. second HDU has index 1).'
    )

    parser.add_argument(
        '--var-hdu', metavar='VAR_HDU', type=str, default=-1,
        help='Index or name of the HDU containing the variance. '
        'Set this to -1 if no variance data is present in the cube. '
        'The default value is %(metavar)s=%(default)s. '
        'NOTE that this value is zero indexed (i.e. third HDU has index 2).'
    )

    parser.add_argument(
        '--mask-hdu', metavar='MASK_HDU', type=str, default=-1,
        help='Index or name of the HDU containing the data mask. '
        'Set this to -1 if no variance data is present in the cube. '
        'The default value is %(metavar)s=%(default)s. '
        'NOTE that this value is zero indexed (i.e. fourth HDU has index 3).'
    )

    parser.add_argument(
        '--outdir', metavar='DIR', type=str, default=None,
        help='Set the directory where extracted spectra will be outputed. '
        'If this parameter is not specified, then a new directory will be '
        'created based on the name of the input cube.'
    )

    parser.add_argument(
        '--wave-ranges', metavar='WAVE_RANGE', type=str, default=None,
        help='Set the wavelength range(s) to stack, in the format '
        'WAVE_RANGE=RANGE1_START-RANGE1_STOP,RANGE2_START-RANGE2_STOP,...'
        ' If not specified, the whole wavelength range is used.'
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    return args


def cube_stack(options=None):
    """
    Run the main program of this module.

    Returns
    -------
    None.

    """
    args = __argshandler(options)

    input_file = args.input_cube[0]

    basename = os.path.basename(args.input_cube[0])
    basename = os.path.splitext(basename)[0]

    if args.outdir is not None:
        basename = os.path.join(args.outdir, basename)

    with fits.open(input_file) as hdul:

        try:
            spec_hdu_index = int(args.spec_hdu)
        except ValueError:
            spec_hdu_index = args.spec_hdu
        else:
            if spec_hdu_index < 0:
                spec_hdu_index = None

        try:
            var_hdu_index = int(args.var_hdu)
        except ValueError:
            var_hdu_index = args.var_hdu
        else:
            if var_hdu_index < 0:
                var_hdu_index = None

        try:
            mask_hdu_index = int(args.mask_hdu)
        except ValueError:
            mask_hdu_index = args.mask_hdu
        else:
            if mask_hdu_index < 0:
                mask_hdu_index = None

        spec_hdu = get_hdu(
            hdul,
            hdu_index=spec_hdu_index,
            valid_names=['data', 'spec', 'spectrum', 'spectra'],
            msg_err_notfound="ERROR: Cannot determine which HDU contains "
                             "spectral data, try to specify it manually!",
            msg_index_error="ERROR: Cannot open HDU {} to read specra!"
        )

        if args.wave_ranges is not None:
            wave_ranges = [
                [float(k) for k in x.split('-')]
                for x in args.wave_ranges.split(',')
            ]
        else:
            wave_ranges = None

        dat, dat_wcs = stack_and_plot(
            spec_hdu,
            basename,
            'data',
            wave_ranges=wave_ranges,
            dpi=args.dpi
        )

        if var_hdu_index is not None:
            var_hdu = get_hdu(
                hdul,
                hdu_index=var_hdu_index,
                valid_names=['stat', 'var', 'variance', 'noise'],
                msg_err_notfound="WARNING: Cannot determine which HDU contains "
                                 "the variance data, try to specify it manually!",
                msg_index_error="ERROR: Cannot open HDU {} to read the "
                                "variance!",
                exit_on_errors=False
            )

            var, var_wcs = stack_and_plot(
                var_hdu,
                basename,
                'variance',
                override_wcs=dat_wcs,
                wave_ranges=wave_ranges,
                dpi=args.dpi
            )

        if mask_hdu_index is not None:
            mask_hdu = get_hdu(
                hdul,
                hdu_index=mask_hdu_index,
                valid_names=['mask', 'platemask', 'footprint', 'dq'],
                msg_err_notfound="WARNING: Cannot determine which HDU contains "
                                 "the mask data, try to specify it manually!",
                msg_index_error="ERROR: Cannot open HDU {} to read the mask!",
                exit_on_errors=False
            )

            mask, mask_wcs = stack_and_plot(
                mask_hdu,
                basename,
                'mask',
                is_mask=True,
                override_wcs=dat_wcs,
                wave_ranges=wave_ranges,
                dpi=args.dpi
            )


if __name__ == '__main__':
    cube_stack()



"""
SPEX - SPectra EXtractor.

Extract spectra from spectral data cubes and find their redshift.

Copyright (C) 2022  Maurizio D'Addona <mauritiusdadd@gmail.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import sys
import time
from urllib import request
import pathlib

from specex.utils import get_pbar

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

TEST_DATA_PATH = os.path.join(pathlib.Path(__file__).parent.resolve(), "data")


def downloadFiles(url_list: tuple, out_dir: str = TEST_DATA_PATH,
                  use_cached: bool = True, report_interval=1):
    """
    Download a set of files with a progressbar.

    Parameters
    ----------
    url_list : tuple
        DESCRIPTION.
    out_dir : str, optional
        DESCRIPTION. The default is TEST_DATA_PATH.
    use_cached : bool, optional
        DESCRIPTION. The default is True.
    report_interval : TYPE, optional
        DESCRIPTION. The default is 1.

    Returns
    -------
    outfile_list : TYPE
        DESCRIPTION.

    """

    def report_pbar(blocks_count, block_size, total_size):
        if (time.perf_counter() - report_pbar.last_time) > report_interval:
            report_pbar.last_time = time.perf_counter()
            downloaded_size = blocks_count * block_size
            progress = downloaded_size / total_size
            pbar = get_pbar(progress)
            report_str = f"\r{pbar} {progress: 6.2%}  "
            report_str += f"{downloaded_size}/{total_size} Bytes\r"
            sys.stdout.write(report_str)
            sys.stdout.flush()
    report_pbar.last_time = time.perf_counter()

    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    outfile_list = []
    for target_url in url_list:
        target_file = os.path.basename(target_url)
        out_file = os.path.join(out_dir, target_file)
        if not (use_cached and os.path.isfile(out_file)):
            print(f"Downloading {target_file}...")
            try:
                package_out_file, headers = request.urlretrieve(
                    target_url,
                    out_file,
                    reporthook=report_pbar
                )
            except Exception as exc:
                print(
                    f"An exception occurred while downloading {target_file}:",
                    str(exc),
                    file=sys.stderr
                )
                continue
        else:
            print(f"Using cached {target_file}...")
        outfile_list.append(os.path.realpath(out_file))
    return outfile_list


def get_muse_test_cube(out_dir: str = TEST_DATA_PATH,
                       use_cached: bool = True):
    TEST_CUBE_URL = 'https://dataportal.eso.org/dataportal_new/file//ADP.2023-09-01T12:56:41.595'
    return downloadFiles(
        [TEST_CUBE_URL,],
        out_dir=out_dir,
        use_cached=use_cached,
        report_interval=10
    )[0]


def get_hst_test_images(out_dir: str = TEST_DATA_PATH,
                        use_cached: bool = True):
    """
    Download an HST test files.

    Parameters
    ----------
    out_dir : str, optional
        The Directory whre to save the files.
        The default is test.TEST_DATA_PATH.
    use_cached : bool, optional
        If True, do not download the file if a cached version exists.
        The default is True.

    Returnsoutfile_list
    -------
    outfile : str
        The absolute path of the downloaded image.
    """
    # See: https://fits.gsfc.nasa.gov/fits_samples.html
    HST_BASE_URL = 'https://fits.gsfc.nasa.gov/samples'

    HST_TEST_FITS = [
        f'{HST_BASE_URL}/NICMOSn4hk12010_mos.fits',
        f'{HST_BASE_URL}/WFPC2ASSNu5780205bx.fits',
        f'{HST_BASE_URL}/WFPC2u5780205r_c0fx.fits',
    ]

    return downloadFiles(
        HST_TEST_FITS,
        out_dir=out_dir,
        use_cached=use_cached,
        report_interval=1
    )
